# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import inspect
import typing
import unittest

import fixtures
import testtools

SKIP_CLASSES = unittest.SkipTest, testtools.TestCase.skipException

SkipException = unittest.SkipTest

SkipTarget = typing.Union[typing.Callable,
                          typing.Type[testtools.TestCase],
                          typing.Type[fixtures.Fixture]]
SkipDecorator = typing.Callable[[SkipTarget], SkipTarget]


SkipOnErrorType = typing.Union[typing.Type[Exception],
                               typing.Tuple[typing.Type[Exception], ...]]


def skip_test(reason: str,
              cause: Exception = None,
              bugzilla: int = None) -> typing.NoReturn:
    """Interrupt test case execution marking it as skipped for given reason"""
    if bugzilla is not None:
        reason += f'\nhttps://bugzilla.redhat.com/show_bug.cgi?id={bugzilla}\n'
    if cause is not None:
        reason += f"\n{cause}\n"
    raise unittest.SkipTest(reason) from cause


def skip(reason: str,
         bugzilla: int = None) -> SkipDecorator:
    """Mark test case for being skipped for a given reason"""
    return _skip_decorator(reason=reason, bugzilla=bugzilla)


def skip_if(reason: str,
            predicate: typing.Callable,
            *args,
            bugzilla: int = None,
            **kwargs) -> \
        SkipDecorator:
    """Mark test case for being skipped for a given reason if it matches"""
    predicate = _get_skip_predicate(predicate, *args, **kwargs)
    return _skip_decorator(reason=reason,
                           unless=False,
                           bugzilla=bugzilla,
                           predicate=predicate)


def skip_unless(reason: str,
                predicate: typing.Callable,
                *args,
                bugzilla: int = None,
                **kwargs) -> \
        SkipDecorator:
    """Mark test case for being skipped for a given reason unless it matches"""
    predicate = _get_skip_predicate(predicate, *args, **kwargs)
    return _skip_decorator(reason=reason,
                           unless=True,
                           bugzilla=bugzilla,
                           predicate=predicate)


def skip_on_error(reason: str,
                  predicate: typing.Callable,
                  *args,
                  error_type: SkipOnErrorType = None,
                  bugzilla: int = None,
                  **kwargs) -> \
        SkipDecorator:
    predicate = _get_skip_predicate(predicate, *args, **kwargs)
    return _skip_decorator(reason=reason,
                           error_type=error_type,
                           bugzilla=bugzilla,
                           predicate=predicate)


def _skip_decorator(reason: str,
                    unless: bool = None,
                    error_type: SkipOnErrorType = None,
                    bugzilla: int = None,
                    predicate: typing.Callable = None) \
        -> SkipDecorator:
    """Mark test case for being skipped for a given reason unless it matches"""

    if error_type is None:
        error_type = tuple()

    def decorator(obj: SkipTarget) -> SkipTarget:
        method = _get_skip_method(obj)

        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            _reason = reason
            cause: typing.Optional[Exception] = None
            if predicate is not None:
                return_value: typing.Any = None
                try:
                    return_value = predicate()
                except error_type as ex:
                    cause = ex
                else:
                    if unless in [None, bool(return_value)]:
                        return method(*args, **kwargs)
                if '{return_value' in reason:
                    _reason = reason.format(return_value=return_value)
                if '{cause' in reason:
                    _reason = reason.format(cause=cause)
            skip_test(reason=_reason, cause=cause, bugzilla=bugzilla)

        if obj is method:
            return wrapper
        else:
            setattr(obj, method.__name__, wrapper)
            return obj

    return decorator


def _get_skip_method(obj: SkipTarget) -> typing.Callable:
    if inspect.isclass(obj):
        assert isinstance(obj, type)
        cls = typing.cast(typing.Type, obj)
        setup_method = getattr(cls, 'setUp', None)
        if callable(setup_method):
            return setup_method
        else:
            # Dummy setUp method implementation for abstract classes
            def setUp(self) -> None:
                super(cls, self).setUp()
            return setUp
    elif callable(obj):
        return obj
    else:
        raise TypeError(f"Object {obj} is not a class or a function")


def _get_skip_predicate(func: typing.Callable, *args, **kwargs) \
        -> typing.Callable:
    if args or kwargs:
        return functools.partial(func, *args, **kwargs)
    else:
        return func



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import io
import os
import tempfile

import tobiko
from tobiko.tests import unit


class ParseIniFileTest(unit.TobikoUnitTest):

    def test_parse_ini_file(self):
        text = """
        my_option = my value
        """
        result = tobiko.parse_ini_file(text=text)
        self.assertEqual({('DEFAULT', 'my_option'): 'my value'},
                         result)

    def test_parse_ini_with_lines(self):
        lines = ['a = 3',
                 'b=4']
        result = tobiko.parse_ini_file(text=lines)
        self.assertEqual({('DEFAULT', 'a'): '3',
                          ('DEFAULT', 'b'): '4'},
                         result)

    def test_parse_ini_with_io(self):
        text = """
        my_option = my value
        """
        fd, temp_file = tempfile.mkstemp(text=True)
        self.addCleanup(os.remove, temp_file)
        with os.fdopen(fd, 'wt') as stream:
            stream.write(text)

        with io.open(temp_file, 'rt') as stream:
            result = tobiko.parse_ini_file(text=stream)
        self.assertEqual({('DEFAULT', 'my_option'): 'my value'},
                         result)

    def test_parse_ini_file_with_default_section(self):
        text = """
        my_option = my value
        """
        result = tobiko.parse_ini_file(text=text, section='default-section')
        self.assertEqual({('default-section', 'my_option'): 'my value'},
                         result)

    def test_parse_ini_file_with_section(self):
        text = """
        [my_section]
        my_option = my value
        """
        result = tobiko.parse_ini_file(text=text)
        self.assertEqual({('my_section', 'my_option'): 'my value'}, result)

    def test_parse_ini_file_empty(self):
        result = tobiko.parse_ini_file(text='')
        self.assertEqual({}, result)

    def test_parse_ini_file_with_commented_option(self):
        text = """
        my_option = 1
        # my_option = 2
        """
        result = tobiko.parse_ini_file(text=text)
        self.assertEqual({('DEFAULT', 'my_option'): '1'},
                         result)

    def test_parse_ini_file_with_duplicate_option(self):
        text = """
        my_option = a
        my_option = b
        """
        result = tobiko.parse_ini_file(text=text)
        self.assertEqual({('DEFAULT', 'my_option'): 'b'},
                         result)



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import subprocess
import sys
import typing

from oslo_log import log

from tobiko.run import _config


LOG = log.getLogger(__name__)


def find_test_files(test_path: typing.Union[str, typing.Iterable[str]] = None,
                    test_filename: str = None,
                    config: _config.RunConfigFixture = None) \
        -> typing.List[str]:
    config = _config.run_confing(config)
    if test_path is None:
        test_path = config.test_path
    elif isinstance(test_path, str):
        test_path = [test_path]
    else:
        test_path = list(test_path)
    if not test_filename:
        test_filename = config.test_filename
    test_files: typing.List[str] = []
    for path in test_path:
        path = os.path.realpath(path)
        if os.path.isfile(path):
            test_files.append(path)
            LOG.debug("Found test file:\n"
                      f"  {path}\n",)
            continue
        if os.path.isdir(path):
            find_dir = path
            find_name = test_filename
        else:
            find_dir = os.path.dirname(path)
            find_name = os.path.basename(path)

        LOG.debug("Find test files...\n"
                  f"  dir: '{find_dir}'\n"
                  f"  name: '{find_name}'")
        try:
            output = subprocess.check_output(
                ['find', find_dir, '-name', find_name],
                universal_newlines=True)
        except subprocess.CalledProcessError as ex:
            LOG.exception("Test files not found.")
            raise FileNotFoundError('Test files not found: \n'
                                    f"  dir: '{find_dir}'\n"
                                    f"  name: '{find_name}'") from ex

        for line in output.splitlines():
            line = line.strip()
            if line:
                test_files.append(line)

        LOG.debug("Found test file(s):\n"
                  "  %s", '  \n'.join(test_files))
    return test_files


def main(test_path: typing.List[str] = None):
    if test_path is None:
        test_path = sys.argv[1:]
    try:
        test_files = find_test_files(test_path=test_path)
    except Exception as ex:
        sys.stderr.write(f'{ex}\n')
        sys.exit(1)
    else:
        output = ''.join(f'{test_file}\n'
                         for test_file in test_files)
        sys.stdout.write(output)
        sys.exit(0)


if __name__ == '__main__':
    main()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks


@keystone.skip_unless_has_keystone_credentials()
class SubnetTest(testtools.TestCase):
    """Tests network creation"""

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    def test_list_subnets(self):
        subnets = neutron.list_subnets()
        subnets_ids = {s['id'] for s in subnets}
        if self.stack.has_ipv4:
            self.assertIn(self.stack.ipv4_subnet_id, subnets_ids)
        if self.stack.has_ipv6:
            self.assertIn(self.stack.ipv6_subnet_id, subnets_ids)

    def test_list_subnet_cidrs(self):
        def _find_prefix(cidr, prefixes):
            for prefix in prefixes:
                if cidr in prefix:
                    return prefix
            return None

        if self.stack.has_ipv4:
            cidr = netaddr.IPNetwork(self.stack.ipv4_subnet_details['cidr'])
            prefixes = [
                netaddr.IPNetwork(prefix)
                for prefix in
                self.stack.subnet_pools_ipv4_stack.subnet_pool['prefixes']]
            self.assertIsNotNone(_find_prefix(cidr, prefixes))
        if self.stack.has_ipv6:
            cidr = netaddr.IPNetwork(self.stack.ipv6_subnet_details['cidr'])
            prefixes = [
                netaddr.IPNetwork(prefix)
                for prefix in
                self.stack.subnet_pools_ipv6_stack.subnet_pool['prefixes']]
            self.assertIsNotNone(_find_prefix(cidr, prefixes))

    def test_get_ipv4_subnet(self):
        if not self.stack.has_ipv4:
            tobiko.skip_test(
                "Stack {self.stack.stack_name} has no IPv4 subnet")
        subnet = neutron.get_subnet(self.stack.ipv4_subnet_id)
        self.assertEqual(self.stack.ipv4_subnet_id, subnet['id'])
        self.assertEqual(self.stack.ipv4_subnet_details, subnet)

    def test_get_ipv6_subnet(self):
        if not self.stack.has_ipv6:
            tobiko.skip_test(
                "Stack {self.stack.stack_name} has no IPv6 subnet")
        subnet = neutron.get_subnet(self.stack.ipv6_subnet_id)
        self.assertEqual(self.stack.ipv6_subnet_id, subnet['id'])
        self.assertEqual(self.stack.ipv6_subnet_details, subnet)

    def test_create_subnet(self):
        network = neutron.create_network(name=self.id())
        subnet_pool_range = "192.168.0.0/16"
        subnet_pool_default_prefixlen = 24
        subnet_pool = neutron.create_subnet_pool(
            name=self.id(),
            prefixes=[subnet_pool_range],
            default_prefixlen=subnet_pool_default_prefixlen)
        subnet = neutron.create_subnet(network=network,
                                       ip_version=4,
                                       subnetpool_id=subnet_pool['id'])
        self.assertEqual(network['id'], subnet['network_id'])
        self.assertIn(netaddr.IPNetwork(subnet['cidr']),
                      netaddr.IPNetwork(subnet_pool_range))
        self.assertEqual(subnet['id'], neutron.get_subnet(subnet=subnet)['id'])
        return subnet

    def test_delete_subnet_with_id(self):
        subnet = self.test_create_subnet()
        neutron.delete_subnet(subnet['id'])
        self.assertRaises(neutron.NoSuchSubnet,
                          neutron.get_subnet, subnet=subnet['id'])

    def test_delete_subnet_with_details(self):
        subnet = self.test_create_subnet()
        neutron.delete_subnet(subnet)
        self.assertRaises(neutron.NoSuchSubnet,
                          neutron.get_subnet, subnet=subnet)

    def test_delete_subnet_with_invalid_id(self):
        self.assertRaises(neutron.NoSuchSubnet, neutron.delete_subnet,
                          '<invalid-id>')



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import


from tobiko.openstack.tests import _neutron
from tobiko.openstack.tests import _nova

InvalidDBConnString = _neutron.InvalidDBConnString
RAFTStatusError = _neutron.RAFTStatusError
test_neutron_agents_are_alive = _neutron.test_neutron_agents_are_alive
test_alive_agents_are_consistent_along_time = (
    _neutron.test_alive_agents_are_consistent_along_time)
test_ovn_dbs_validations = _neutron.test_ovn_dbs_validations
test_ovs_bridges_mac_table_size = _neutron.test_ovs_bridges_mac_table_size
test_ovs_namespaces_are_absent = _neutron.test_ovs_namespaces_are_absent
test_ovs_interfaces_are_absent = _neutron.test_ovs_interfaces_are_absent
test_raft_cluster = _neutron.test_raft_cluster
test_raft_clients_connected = _neutron.test_raft_clients_connected

test_evacuable_server_creation = _nova.test_evacuable_server_creation
test_server_creation = _nova.test_server_creation
test_servers_creation = _nova.test_servers_creation
test_server_creation_and_shutoff = _nova.test_server_creation_and_shutoff
test_server_creation_no_fip = _nova.test_server_creation_no_fip
TestServerCreationStack = _nova.TestServerCreationStack
TestEvacuableServerCreationStack = _nova.TestEvacuableServerCreationStack
test_ovsdb_transactions = _neutron.test_ovsdb_transactions



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import os
import typing

from oslo_log import log

import tobiko
from tobiko.shell import sh
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


class CurlProcessFixture(tobiko.SharedFixture):

    _command: sh.ShellCommand
    _process: sh.ShellProcessFixture
    _result: sh.ShellExecuteResult

    def __init__(self,
                 url: str,
                 continue_at: int = None,
                 create_dirs: bool = None,
                 file_name: str = None,
                 head_only: bool = None,
                 headers_file_name: str = None,
                 location: bool = None,
                 max_redirs: int = None,
                 ssh_client: ssh.SSHClientType = None,
                 sudo: bool = None):
        super().__init__()
        tobiko.check_valid_type(continue_at, int, type(None))
        tobiko.check_valid_type(create_dirs, bool, type(None))
        tobiko.check_valid_type(head_only, bool, type(None))
        tobiko.check_valid_type(location, bool, type(None))
        tobiko.check_valid_type(max_redirs, int, type(None))
        tobiko.check_valid_type(file_name, str, type(None))
        tobiko.check_valid_type(headers_file_name, str, type(None))
        tobiko.check_valid_type(ssh_client, ssh.SSHClientFixture, bool,
                                type(None))
        tobiko.check_valid_type(sudo, bool, type(None))
        tobiko.check_valid_type(url, str)
        self.continue_at = continue_at
        self.create_dirs = create_dirs
        self.head_only = head_only
        self.headers_file_name = headers_file_name
        self.location = location
        self.max_redirs = max_redirs
        self.file_name = file_name
        self.ssh_client = ssh_client
        self.sudo = sudo
        self.url = url

    @property
    def command(self) -> sh.ShellCommand:
        try:
            return self._command
        except AttributeError:
            pass
        command = sh.shell_command('curl') + self.get_options() + self.url
        self._command = command
        self.addCleanup(self._del_command)
        return command

    def _del_command(self):
        try:
            del self._command
        except AttributeError:
            pass

    @property
    def process(self) -> sh.ShellProcessFixture:
        try:
            return self._process
        except AttributeError:
            raise RuntimeError("Curl process not stated") from None

    def setup_fixture(self):
        if self.create_dirs and self.headers_file_name is not None:
            headers_dir = os.path.dirname(self.headers_file_name)
            sh.execute(f'mkdir -p "{headers_dir}"',
                       ssh_client=self.ssh_client,
                       sudo=self.sudo)
        process = sh.process(self.command,
                             ssh_client=self.ssh_client,
                             sudo=self.sudo)
        self._process = self.useFixture(process)

    def start(self) -> 'CurlProcessFixture':
        return tobiko.setup_fixture(self)

    def stop(self) -> 'CurlProcessFixture':
        try:
            process = self._process
        except RuntimeError:
            pass  # process not started
        else:
            process.kill(sudo=self.sudo)
        return tobiko.cleanup_fixture(self)

    def execute(self,
                retry_count: int = None,
                retry_timeout: tobiko.Seconds = None,
                retry_interval: tobiko.Seconds = None) -> \
            sh.ShellExecuteResult:
        for attempt in tobiko.retry(count=retry_count,
                                    timeout=retry_timeout,
                                    interval=retry_interval,
                                    default_count=1):
            self.start()
            result = self.wait(check=attempt.is_last)
            if result.exit_status == 0:
                break
            self.stop()
        else:
            raise RuntimeError("Retry loop broken")
        return result

    def wait(self,
             check: bool = False,
             expect_exit_status: int = None) \
            -> sh.ShellExecuteResult:
        if expect_exit_status is None and check:
            expect_exit_status = 0
        self._result = result = sh.execute_process(
            self.process, expect_exit_status=expect_exit_status)
        return result

    @property
    def result(self) -> sh.ShellExecuteResult:
        if self.executed:
            return self._result
        else:
            raise RuntimeError("Process not terminated") from None

    @property
    def executed(self) -> bool:
        return hasattr(self, '_result')

    def get_options(self) -> sh.ShellCommand:
        options = sh.ShellCommand()
        if self.continue_at:
            if self.continue_at < 0:
                options += "-C -"
            else:
                options += f"-C '{self.continue_at}'"
        if self.create_dirs:
            options += '--create-dirs'
        if self.head_only:
            options += '-I'
        if self.headers_file_name:
            options += f'-D "{self.headers_file_name}"'
        if self.location:
            options += '-L'
        if self.max_redirs is not None:
            options += f"--max-redirs '{self.max_redirs}'"
        if self.file_name:
            options += f"-o '{self.file_name}'"
        return options

    def open_output_file(self):
        tobiko.check_valid_type(self.file_name, str)
        assert self.file_name


def curl_process(url: str,
                 continue_at: int = None,
                 create_dirs: bool = None,
                 head_only: bool = None,
                 headers_file_name: str = None,
                 location: bool = None,
                 max_redirs: int = None,
                 file_name: str = None,
                 ssh_client: ssh.SSHClientType = None,
                 sudo: bool = None) \
        -> CurlProcessFixture:
    return CurlProcessFixture(url=url,
                              continue_at=continue_at,
                              create_dirs=create_dirs,
                              head_only=head_only,
                              headers_file_name=headers_file_name,
                              location=location,
                              max_redirs=max_redirs,
                              file_name=file_name,
                              ssh_client=ssh_client,
                              sudo=sudo)


class CurlHeader(collections.UserDict):

    def __init__(self,
                 head_line: typing.Optional[str],
                 *args,
                 **kwargs):
        self.head_line = head_line
        super(CurlHeader, self).__init__(*args, **kwargs)

    @property
    def content_length(self) -> typing.Optional[int]:
        length = self.get('content-length')
        if length is None:
            return None
        else:
            return int(length)

    def __repr__(self):
        entries = super().__repr__()
        return f"CurlHeader({self.head_line!r}, {entries})"


def get_url_header(url: str,
                   location: typing.Optional[bool] = True,
                   max_redirs: int = None,
                   ssh_client: ssh.SSHClientType = None,
                   retry_timeout: tobiko.Seconds = None,
                   retry_interval: tobiko.Seconds = None,
                   retry_count: int = None,
                   **process_params) -> CurlHeader:
    headers = list_url_headers(url=url,
                               location=location,
                               max_redirs=max_redirs,
                               ssh_client=ssh_client,
                               retry_timeout=retry_timeout,
                               retry_interval=retry_interval,
                               retry_count=retry_count,
                               **process_params)
    return headers[-1]


def list_url_headers(url: str,
                     location: typing.Optional[bool] = True,
                     max_redirs: int = None,
                     ssh_client: ssh.SSHClientType = None,
                     retry_timeout: tobiko.Seconds = None,
                     retry_interval: tobiko.Seconds = None,
                     retry_count: int = None,
                     sudo: bool = None) \
        -> tobiko.Selection[CurlHeader]:
    process = curl_process(url=url,
                           location=location,
                           head_only=True,
                           max_redirs=max_redirs,
                           ssh_client=ssh_client,
                           sudo=sudo)
    result = process.execute(retry_count=retry_count,
                             retry_timeout=retry_timeout,
                             retry_interval=retry_interval)
    return parse_headers(result.stdout)


def read_headers_file(headers_file_name: str,
                      ssh_client: ssh.SSHClientType = None,
                      sudo: bool = None) \
        -> tobiko.Selection[CurlHeader]:
    result = sh.execute(f'cat "{headers_file_name}"',
                        ssh_client=ssh_client,
                        sudo=sudo)
    return parse_headers(result.stdout)


def parse_headers(headers_text: str) \
        -> tobiko.Selection[CurlHeader]:
    headers = tobiko.Selection[CurlHeader]()
    entries: typing.Dict[str, str] = {}
    line_number = 0
    head_line: typing.Optional[str] = None
    for line in headers_text.splitlines():
        line = line.strip()
        line_number += 1
        if not line:
            if entries:
                headers.append(CurlHeader(head_line, entries))
                head_line = None
                entries = {}
        elif ':' in line:
            key, value = line.split(':', 1)
            entries[key.lower().strip()] = value.strip()
        elif len(entries) == 0:
            head_line = line
        else:
            raise ValueError(f'Invalid line {line_number} in header: {line}\n"'
                             f'"{headers_text}')
    if entries:
        headers.append(CurlHeader(head_line, entries))
    return headers


def default_download_dir() -> str:
    return '~/.tobiko/download'


def download_file(url: str,
                  file_name: str = None,
                  cached: bool = False,
                  check: bool = True,
                  download_dir: str = None,
                  continue_at: int = None,
                  create_dirs: bool = None,
                  headers_file_name: str = None,
                  location: typing.Optional[bool] = True,
                  max_redirs: int = None,
                  wait: bool = True,
                  retry_count: int = None,
                  retry_timeout: tobiko.Seconds = None,
                  retry_interval: tobiko.Seconds = None,
                  ssh_client: ssh.SSHClientType = None,
                  sudo: bool = None) \
        -> CurlProcessFixture:
    if file_name is None:
        if download_dir is None:
            download_dir = default_download_dir()
            create_dirs = True
        file_name = os.path.join(download_dir,
                                 os.path.basename(url))
    if headers_file_name is None:
        headers_file_name = file_name + '.headers'
        create_dirs = True

    process = curl_process(url=url,
                           file_name=file_name,
                           headers_file_name=headers_file_name,
                           continue_at=continue_at,
                           create_dirs=create_dirs,
                           location=location,
                           max_redirs=max_redirs,
                           ssh_client=ssh_client,
                           sudo=sudo)
    if cached:
        try:
            assert_downloaded_file(file_name=file_name,
                                   headers_file_name=headers_file_name,
                                   ssh_client=ssh_client,
                                   sudo=sudo)
        except tobiko.FailureException:  # type: ignore
            pass
        else:
            LOG.debug(f"File '{url}' already downloaded.")
            return process

    LOG.debug(f"Downloading file ('{url}' -> '{file_name}')...")
    if not wait:
        process.start()
        return process

    process.execute(retry_count=retry_count,
                    retry_timeout=retry_timeout,
                    retry_interval=retry_interval)
    LOG.debug(f"File download complete ('{url}' -> '{file_name}').")
    if check:
        assert_downloaded_file(file_name=file_name,
                               headers_file_name=headers_file_name,
                               ssh_client=ssh_client,
                               sudo=sudo)
    return process


def assert_downloaded_file(file_name: str,
                           headers_file_name: str,
                           ssh_client: ssh.SSHClientType = None,
                           sudo: bool = None):
    try:
        header = read_headers_file(headers_file_name=headers_file_name,
                                   ssh_client=ssh_client,
                                   sudo=sudo)[-1]
    except sh.ShellCommandFailed as ex:
        tobiko.fail(f"Error reading headers file '{headers_file_name}': {ex}")
    else:
        file_size = header.content_length
        if file_size is not None:
            sh.assert_file_size(file_size=header.content_length,
                                file_name=file_name,
                                ssh_client=ssh_client,
                                sudo=sudo)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.openstack.designate import _client
from tobiko.openstack.designate import _constants

DesignateClientFixture = _client.DesignateClientFixture
designate_client = _client.designate_client
get_designate_client = _client.get_designate_client
DESIGNATE_CLIENT_CLASSES = _client.DESIGNATE_CLIENT_CLASSES

designate_zone_id = _client.designate_zone_id
get_designate_zone = _client.get_designate_zone
list_recordsets = _client.list_recordsets
create_recordsets = _client.create_recordsets
get_recordset = _client.get_recordset

# Waiters
wait_for_status = _client.wait_for_status

# Constants
STATUS = _constants.STATUS
ACTIVE = _constants.ACTIVE
ERROR = _constants.ERROR



# Copyright (c) 2023 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack.stacks._neutron import StatelessSecurityGroupFixture

DESCRIPTION = "Amazing Stateless Sec Group number {i}"
NUM_SEC_GROUPS = 10


@keystone.skip_unless_has_keystone_credentials()
class StatelessSecurityGroupTest(testtools.TestCase):
    """Tests Stateless Security Group creation"""

    stack_list = [tobiko.required_fixture(
                      StatelessSecurityGroupFixture,
                      fixture_id=str(i),
                      description=DESCRIPTION.format(i=i))
                  for i in range(NUM_SEC_GROUPS)]
    ssg_fixture_list: list = []

    @classmethod
    def tearDownClass(cls):
        for stack in cls.stack_list:
            tobiko.cleanup_fixture(stack.fixture)

    @classmethod
    def setUpClass(cls):
        for stack in cls.stack_list:
            cls.ssg_fixture_list.append(tobiko.use_fixture(stack.fixture))

    def test_stateless_sec_group_list_find(self):
        self.assertEqual(NUM_SEC_GROUPS, len(self.ssg_fixture_list))
        for i, ssg_fixture in enumerate(self.ssg_fixture_list):
            ssg_name = (f"{StatelessSecurityGroupFixture.__module__}."
                        f"{StatelessSecurityGroupFixture.__qualname__}-{i}")
            self.assertEqual(ssg_name, ssg_fixture.name)
            ssg = neutron.list_security_groups(name=ssg_name).unique
            self.assertCountEqual(ssg.keys(),
                                  ssg_fixture.security_group.keys())
            for k in ssg.keys():
                if k != 'security_group_rules':
                    self.assertEqual(ssg[k], ssg_fixture.security_group[k])
                else:
                    # the elements from the lists ssg['security_group_rules']
                    # and ssg_fixture.security_group['security_group_rules']
                    # are equal, but they could be ordered in a different way
                    self.assertCountEqual(ssg[k],
                                          ssg_fixture.security_group[k])

    def test_stateless_sec_group_list_parameters(self):
        for i, ssg_fixture in enumerate(self.ssg_fixture_list):
            self.assertEqual(DESCRIPTION.format(i=i),
                             ssg_fixture.security_group['description'])



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import typing

import netaddr

import tobiko
from tobiko.openstack.neutron import _client
from tobiko.openstack.neutron import _network
from tobiko.openstack.neutron import _subnet
from tobiko.shell import ssh
from tobiko.shell import ping


PortType = typing.Dict[str, typing.Any]
PortIdType = typing.Union[str, PortType]


def get_port_id(port: PortIdType) -> str:
    if isinstance(port, str):
        return port
    else:
        return port['id']


def get_port(port: PortIdType,
             client: _client.NeutronClientType = None,
             **params) -> PortType:
    port_id = get_port_id(port)
    try:
        return _client.neutron_client(client).show_port(
            port_id, **params)['port']
    except _client.NotFound as ex:
        raise NoSuchPort(id=port_id) from ex


def create_port(client: _client.NeutronClientType = None,
                network: _network.NetworkIdType = None,
                add_cleanup=True,
                **params) -> PortType:
    if 'network_id' not in params:
        if network is None:
            from tobiko.openstack import stacks
            network_id = tobiko.setup_fixture(
                stacks.NetworkStackFixture).network_id
        else:
            network_id = _network.get_network_id(network)
        params['network_id'] = network_id
    port = _client.neutron_client(client).create_port(
        body={'port': params})['port']
    if add_cleanup:
        tobiko.add_cleanup(cleanup_port, port=port, client=client)
    return port


def cleanup_port(port: PortIdType,
                 client: _client.NeutronClientType = None):
    try:
        delete_port(port=port, client=client)
    except NoSuchPort:
        pass


def update_port(port: PortIdType,
                client: _client.NeutronClientType = None,
                **params) -> PortType:
    port_id = get_port_id(port)
    reply = _client.neutron_client(client).update_port(port_id,
                                                       body={'port': params})
    return reply['port']


def delete_port(port: PortIdType,
                client: _client.NeutronClientType = None):
    port_id = get_port_id(port)
    try:
        _client.neutron_client(client).delete_port(port_id)
    except _client.NotFound as ex:
        raise NoSuchPort(id=port_id) from ex


DeviceIdType = typing.Union[str, typing.Any]


def get_device_id(device: DeviceIdType) -> str:
    if isinstance(device, str):
        return device
    elif isinstance(device, abc.Mapping):
        return device['id']
    elif hasattr(device, 'id'):
        return getattr(device, 'id')
    else:
        raise TypeError(f'{device!r} is not a valid device type')


def list_ports(client: _client.NeutronClientType = None,
               device: DeviceIdType = None,
               network: _network.NetworkIdType = None,
               subnet: _subnet.SubnetIdType = None,
               **params) -> tobiko.Selection[PortType]:
    if device is not None:
        params.setdefault('device_id', get_device_id(device))
    if network is not None:
        params.setdefault('network_id', _network.get_network_id(network))
    if subnet is not None:
        subnet_id = _subnet.get_subnet_id(subnet)
        params.setdefault('fixed_ips', f'subnet_id={subnet_id}')
    ports = _client.neutron_client(client).list_ports(**params)['ports']
    return tobiko.select(ports)


def find_port(client: _client.NeutronClientType = None,
              unique=False,
              default: PortType = None,
              **params):
    """Look for a port matching some property values"""
    ports = list_ports(client=client, **params)
    if default is None or ports:
        if unique:
            return ports.unique
        else:
            return ports.first
    else:
        return default


def list_port_ip_addresses(port: PortType,
                           subnet: _subnet.SubnetIdType = None,
                           ip_version: int = None,
                           check_connectivity: bool = False,
                           ssh_client: ssh.SSHClientFixture = None) -> \
        tobiko.Selection[netaddr.IPAddress]:
    if subnet is not None:
        subnet = _subnet.get_subnet_id(subnet)
    addresses = tobiko.Selection[netaddr.IPAddress](
        netaddr.IPAddress(fixed_ip['ip_address'])
        for fixed_ip in port['fixed_ips']
        if subnet is None or subnet == fixed_ip['subnet_id'])
    if ip_version is not None:
        addresses = addresses.with_attributes(version=ip_version)
    if addresses and check_connectivity:
        hosts = ping.list_reachable_hosts(addresses, ssh_client=ssh_client)
        addresses = tobiko.Selection(netaddr.IPAddress(host) for host in hosts)
    return addresses


def find_port_ip_address(port: PortType,
                         subnet: _subnet.SubnetIdType = None,
                         ip_version: int = None,
                         check_connectivity: bool = False,
                         ssh_client: ssh.SSHClientFixture = None,
                         unique: bool = False) -> netaddr.IPAddress:
    addresses = list_port_ip_addresses(port=port,
                                       subnet=subnet,
                                       ip_version=ip_version,
                                       check_connectivity=check_connectivity,
                                       ssh_client=ssh_client)
    if unique:
        return addresses.unique
    else:
        return addresses.first


def list_device_ip_addresses(device: DeviceIdType,
                             network: _network.NetworkIdType = None,
                             ip_version: int = None,
                             check_connectivity: bool = False,
                             ssh_client: ssh.SSHClientFixture = None,
                             need_dhcp: bool = None,
                             client: _client.NeutronClientType = None,
                             **subnet_params) -> \
        tobiko.Selection[netaddr.IPAddress]:
    ports = list_ports(device=device,
                       network=network,
                       client=client)
    if need_dhcp is not None:
        subnet_params['enable_dhcp'] = bool(need_dhcp)
    subnets = _subnet.list_subnets(network=network,
                                   ip_version=ip_version,
                                   client=client,
                                   **subnet_params)
    addresses = tobiko.Selection[netaddr.IPAddress](
        port_ip
        for subnet in subnets
        for port in ports
        for port_ip in list_port_ip_addresses(port=port,
                                              subnet=subnet,
                                              ip_version=ip_version))
    if addresses and check_connectivity:
        hosts = ping.list_reachable_hosts(addresses, ssh_client=ssh_client)
        addresses = tobiko.Selection(netaddr.IPAddress(host) for host in hosts)
    return addresses


def find_device_ip_address(device: DeviceIdType,
                           network: _network.NetworkIdType,
                           ip_version: int = None,
                           check_connectivity: bool = False,
                           ssh_client: ssh.SSHClientFixture = None,
                           need_dhcp: bool = None,
                           client: _client.NeutronClientType = None,
                           unique: bool = False,
                           **subnet_params) -> netaddr.IPAddress:
    addresses = list_device_ip_addresses(device=device,
                                         network=network,
                                         ip_version=ip_version,
                                         check_connectivity=check_connectivity,
                                         ssh_client=ssh_client,
                                         need_dhcp=need_dhcp,
                                         client=client,
                                         **subnet_params)
    if unique:
        return addresses.unique
    else:
        return addresses.first


class NoSuchPort(tobiko.ObjectNotFound):
    message = "No such port found for {id!r}"



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr
import pytest
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.shell import ping
from tobiko.shell import sh


@keystone.skip_unless_has_keystone_credentials()
class CirrosServerStackTest(testtools.TestCase):
    """Tests connectivity to Nova instances via floating IPs"""

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    nameservers_filenames: typing.Optional[typing.Sequence[str]] = None

    @property
    def peer_ssh_client(self):
        return self.stack.ssh_client

    def test_ping_floating_ip(self):
        """Test connectivity to floating IP address"""
        ping.assert_reachable_hosts([self.stack.floating_ip_address])

    def test_ping_fixed_ipv4(self):
        ping.assert_reachable_hosts([self.get_fixed_ip(ip_version=4)],
                                    ssh_client=self.peer_ssh_client)

    def test_ping_fixed_ipv6(self):
        ping.assert_reachable_hosts([self.get_fixed_ip(ip_version=6)],
                                    ssh_client=self.peer_ssh_client)

    def get_fixed_ip(self, ip_version: int):
        try:
            return self.stack.find_fixed_ip(ip_version=ip_version)
        except tobiko.ObjectNotFound:
            self.skipTest(f"Server {self.stack.server_id} has any "
                          f"IPv{ip_version} address.")

    def test_ssh_connect(self):
        """Test SSH connectivity via Paramiko SSHClient"""
        self.stack.ssh_client.connect()

    def test_hostname(self):
        """Test that hostname of instance server matches Nova server name"""
        self.stack.ssh_client.connect()

        stdout = sh.execute('hostname',
                            ssh_client=self.stack.ssh_client).stdout
        hostname = stdout.strip().split('.', 1)[0]
        self.assertEqual(self.stack.server_name, hostname)

    def test_console_output(self):
        # wait for server to be ready for connection
        self.stack.ssh_client.connect()
        output = self.stack.console_output
        self.assertTrue(output)

    def test_swap_file(self):
        if getattr(self.stack, 'swap_maxsize', None) is None:
            self.skipTest('Swap maxsize is None')

        cloud_config = self.stack.cloud_config
        self.assertEqual({'filename': self.stack.swap_filename,
                          'size': self.stack.swap_size or "auto",
                          'maxsize': self.stack.swap_maxsize},
                         cloud_config['swap'])

        nova.wait_for_cloud_init_done(ssh_client=self.stack.ssh_client)
        # check swap file exists
        sh.execute(f"ls -lh '{self.stack.swap_filename}'",
                   ssh_client=self.stack.ssh_client)
        # check swap file is mounted
        swaps_table = sh.execute("cat /proc/swaps",
                                 ssh_client=self.stack.ssh_client).stdout
        mounted_filenames = [line.split()[0]
                             for line in swaps_table.splitlines()[1:]]
        self.assertIn(self.stack.swap_filename, mounted_filenames, swaps_table)

    def test_ipv4_subnet_nameservers(self):
        self._test_subnet_nameservers(ip_version=4)

    def test_ipv6_subnet_nameservers(self):
        self._test_subnet_nameservers(ip_version=6)

    def _test_subnet_nameservers(self, ip_version):
        def get_subnet_nameservers(network_stack, ip_version):
            # returns a tuple with:
            # 1. the subnet name (str)
            # 2. the subnet dns nameservers (list of str)
            subnet_details = (network_stack.ipv4_subnet_details
                              if ip_version == 4
                              else network_stack.ipv6_subnet_details)
            nameservers = [netaddr.IPAddress(nameserver)
                           for nameserver in subnet_details['dns_nameservers']]
            return (subnet_details['name'], nameservers)

        subnet_name, subnet_nameservers = get_subnet_nameservers(
            self.stack.network_stack, ip_version)

        if self.stack.has_vlan:
            vlan_subnet_name, vlan_subnet_nameservers = get_subnet_nameservers(
                self.stack.vlan_network_stack, ip_version)
            subnet_names = ','.join([subnet_name, vlan_subnet_name])
            subnet_nameservers += vlan_subnet_nameservers
        else:
            subnet_names = subnet_name

        if not subnet_nameservers:
            self.skipTest(f"Subnet(s) '{subnet_names}' have no IPv{ip_version}"
                          " nameservers")
        server_nameservers = sh.list_nameservers(
            ssh_client=self.stack.ssh_client, ip_version=ip_version,
            filenames=self.nameservers_filenames)
        self.assertEqual(subnet_nameservers, server_nameservers)


class EvacuablesServerStackTest(CirrosServerStackTest):

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.EvacuableServerStackFixture)

    def test_image_fixture_tags(self):
        image_fixture = self.stack.image_fixture
        self.assertEqual(['evacuable'], image_fixture.tags)

    def test_image_tags(self):
        image = self.stack.image_fixture.get_image()
        self.assertEqual(['evacuable'], image.tags)


@pytest.mark.flaky(reruns=2, reruns_delay=60)
class CirrosPeerServerStackTest(CirrosServerStackTest):

    #: Stack of resources with an HTTP server
    stack = tobiko.required_fixture(stacks.CirrosPeerServerStackFixture)

    @property
    def peer_ssh_client(self):
        return self.stack.peer_stack.ssh_client

    def test_ping_floating_ip(self):
        self.skipTest(f"Server '{self.stack.server_id}' has any floating IP")



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import metalsmith
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import _client


LOG = log.getLogger(__name__)


CLIENT_CLASSES = (metalsmith.Provisioner,)
MetalsmithClient = typing.Union[metalsmith.Provisioner]


class MetalsmithClientFixture(_client.OpenstackClientFixture):

    def init_client(self, session) -> MetalsmithClient:
        return metalsmith.Provisioner(session=session)


class MetalsmithClientManager(_client.OpenstackClientManager):

    def create_client(self, session) -> MetalsmithClientFixture:
        return MetalsmithClientFixture(session=session)


CLIENTS = MetalsmithClientManager()


def metalsmith_client_manager(manager: MetalsmithClientManager = None) \
        -> MetalsmithClientManager:
    if manager is None:
        manager = CLIENTS
    return manager


MetalsmithClientType = typing.Union[
    MetalsmithClient,
    MetalsmithClientFixture,
    typing.Type[MetalsmithClientFixture]]


def metalsmith_client(obj: MetalsmithClientType = None) \
        -> MetalsmithClient:
    if obj is None:
        return get_metalsmith_client()

    if isinstance(obj, CLIENT_CLASSES):
        return obj

    fixture = tobiko.setup_fixture(obj)
    if isinstance(fixture, MetalsmithClientFixture):
        assert fixture.client is not None
        return fixture.client

    message = f"Object '{obj}' is not a MetalsmithProvisionerFixture"
    raise TypeError(message)


def get_metalsmith_client(session: keystone.KeystoneSessionType = None,
                          shared=True,
                          init_client=None,
                          manager: MetalsmithClientManager = None) \
        -> MetalsmithClient:
    manager = metalsmith_client_manager(manager)
    client = manager.get_client(session=session,
                                shared=shared,
                                init_client=init_client)
    tobiko.setup_fixture(client)
    return client.client



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import netaddr
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.shell import ping


class KeyPairTest(testtools.TestCase):

    stack = tobiko.required_fixture(stacks.KeyPairStackFixture)

    def test_key_files(self):
        self.assertTrue(os.path.isfile(self.stack.key_file))
        self.assertTrue(os.path.isfile(self.stack.key_file + '.pub'))


@keystone.skip_unless_has_keystone_credentials()
class ClientTest(testtools.TestCase):

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    @nova.skip_if_missing_hypervisors(count=1)
    def test_list_hypervisors(self):
        hypervisor = nova.list_hypervisors().first
        self.assertIsInstance(hypervisor.id, str)
        self.assertTrue(hypervisor.hypervisor_hostname)
        netaddr.IPAddress(hypervisor.host_ip)

    @nova.skip_if_missing_hypervisors(count=1)
    def test_list_hypervisors_without_details(self):
        hypervisor = nova.list_hypervisors(detailed=False).first
        self.assertIsInstance(hypervisor.id, str)
        self.assertTrue(hypervisor.hypervisor_hostname)
        self.assertFalse(hasattr(hypervisor, 'host_ip'))

    @nova.skip_if_missing_hypervisors(count=1)
    def test_list_hypervisors_with_hypervisor_hostname(self):
        hypervisor1 = nova.list_hypervisors().first
        hypervisor2 = nova.list_hypervisors(
            hypervisor_hostname=hypervisor1.hypervisor_hostname).unique
        self.assertEqual(hypervisor1, hypervisor2)

    @nova.skip_if_missing_hypervisors(count=1)
    def test_find_hypervisor(self):
        hypervisor = nova.find_hypervisor()
        self.assertIsNotNone(hypervisor)

    @nova.skip_if_missing_hypervisors(count=2)
    def test_find_hypervisor_with_unique(self):
        self.assertRaises(tobiko.MultipleObjectsFound, nova.find_hypervisor,
                          unique=True)

    @nova.skip_if_missing_hypervisors(count=2)
    def test_find_hypervisor_without_unique(self):
        hypervisor = nova.find_hypervisor()
        self.assertIsNotNone(hypervisor)

    def test_get_console_output(self):
        nova.activate_server(self.stack.server_id)
        output = nova.get_console_output(server=self.stack.server_id,
                                         length=50,
                                         timeout=60.)
        self.assertTrue(output)

    def test_list_servers(self):
        server_id = self.stack.server_id
        for server in nova.list_servers():
            if server_id == server.id:
                break
        else:
            self.fail('Server {} not found'.format(server_id))

    def test_find_server(self):
        server_id = self.stack.server_id
        server = nova.find_server(id=server_id, unique=True)
        self.assertEqual(server_id, server.id)

    def test_wait_for_server_status(self):
        server_id = self.stack.server_id
        server = nova.wait_for_server_status(server=server_id, status='ACTIVE')
        self.assertEqual(server_id, server.id)
        self.assertEqual('ACTIVE', server.status)


class ServerActionsStack(stacks.CirrosServerStackFixture):
    pass


class ServerActionsTest(testtools.TestCase):

    stack = tobiko.required_fixture(ServerActionsStack)

    def test_activate_server(self, initial_status='SHUTOFF'):
        self.stack.ensure_server_status(initial_status)
        server = nova.activate_server(self.stack.server_id)
        self.assertEqual('ACTIVE', server.status)
        ping.assert_reachable_hosts([self.stack.ip_address])

    def test_activate_server_when_shutoff(self):
        self.test_activate_server(initial_status='SHUTOFF')

    def test_shutoff_server(self, initial_status='ACTIVE'):
        self.stack.ensure_server_status(initial_status)
        server = nova.shutoff_server(self.stack.server_id)
        self.assertEqual('SHUTOFF', server.status)
        ping.assert_unreachable_hosts([self.stack.ip_address])

    def test_shutoff_server_when_shutoff(self):
        self.test_shutoff_server(initial_status='SHUTOFF')


@keystone.skip_unless_has_keystone_credentials()
class HypervisorTest(testtools.TestCase):

    def test_skip_if_missing_hypervisors(self, count=1, should_skip=False,
                                         **params):
        if should_skip:
            expected_exeption = self.skipException
        else:
            expected_exeption = self.failureException

        @nova.skip_if_missing_hypervisors(count=count, **params)
        def method():
            raise self.fail('Not skipped')

        exception = self.assertRaises(expected_exeption, method)
        if should_skip:
            hypervisors = nova.list_hypervisors(**params)
            message = "missing {!r} hypervisor(s)".format(
                count - len(hypervisors))
            if params:
                message += " with {!s}".format(
                    ','.join('{!s}={!r}'.format(k, v)
                             for k, v in params.items()))
            self.assertEqual(message, str(exception))
        else:
            self.assertEqual('Not skipped', str(exception))

    def test_skip_if_missing_hypervisors_with_no_hypervisors(self):
        self.test_skip_if_missing_hypervisors(id=-1,
                                              should_skip=True)

    def test_skip_if_missing_hypervisors_with_big_count(self):
        self.test_skip_if_missing_hypervisors(count=1000000,
                                              should_skip=True)


@keystone.skip_unless_has_keystone_credentials()
class ServiceTest(testtools.TestCase):

    def test_wait_for_services_up(self):
        nova.wait_for_services_up()


class MigrateServerStack(stacks.CirrosServerStackFixture):
    pass


@keystone.skip_unless_has_keystone_credentials()
@nova.skip_if_missing_hypervisors(count=2)
class MigrateServerTest(testtools.TestCase):

    stack = tobiko.required_fixture(MigrateServerStack)

    def test_migrate_server(self):
        """Tests cold migration actually changes hypervisor
        """
        server = self.setup_server()
        initial_hypervisor = nova.get_server_hypervisor(server)

        server = self.migrate_server(server)

        final_hypervisor = nova.get_server_hypervisor(server)
        self.assertNotEqual(initial_hypervisor, final_hypervisor)

    def test_migrate_server_with_host(self):
        """Tests cold migration actually ends on target hypervisor
        """
        server = self.setup_server()
        initial_hypervisor = nova.get_server_hypervisor(server)
        for hypervisor in nova.list_hypervisors(status='enabled', state='up'):
            if initial_hypervisor != hypervisor.hypervisor_hostname:
                target_hypervisor = hypervisor.hypervisor_hostname
                break
        else:
            tobiko.skip_test("Cannot find a valid hypervisor host to migrate "
                             "server to")

        server = self.migrate_server(server=server, host=target_hypervisor)

        final_hypervisor = nova.get_server_hypervisor(server)
        self.assertEqual(target_hypervisor, final_hypervisor)

    def setup_server(self):
        server = self.stack.ensure_server_status('ACTIVE')
        self.assertEqual('ACTIVE', server.status)
        return server

    def migrate_server(self, server, **params):
        self.assertEqual('ACTIVE', server.status)
        nova.migrate_server(server, **params)

        server = nova.wait_for_server_status(server, 'VERIFY_RESIZE')
        self.assertEqual('VERIFY_RESIZE', server.status)
        nova.confirm_resize(server)

        server = nova.wait_for_server_status(
            server, 'ACTIVE', transient_status={'VERIFY_RESIZE'})
        self.assertEqual('ACTIVE', server.status)

        ping.ping_until_received(self.stack.ip_address).assert_replied()
        return server



"""Exported method Container.start()."""

from __future__ import absolute_import

import logging
import os
import select
import signal
import socket
import sys
import termios
import tty

CONMON_BUFSZ = 8192


class Mixin:
    """Publish start() for inclusion in Container class."""

    def start(self):
        """Start container, return container on success.

        Will block if container has been detached.
        """
        # pylint: disable=protected-access
        with self._client() as podman:
            logging.debug('Starting Container "%s"', self._id)
            results = podman.StartContainer(self._id)
            logging.debug('Started Container "%s"', results['container'])

            if not hasattr(self, 'pseudo_tty') or self.pseudo_tty is None:
                return self._refresh(podman)

            logging.debug('Setting up PseudoTTY for Container "%s"',
                          results['container'])

            try:
                # save off the old settings for terminal
                tcoldattr = termios.tcgetattr(self.pseudo_tty.stdin)
                tty.setraw(self.pseudo_tty.stdin)

                # initialize container's window size
                self.resize_handler(None, sys._getframe(0))

                # catch any resizing events and send the resize info
                # to the control fifo "socket"
                signal.signal(signal.SIGWINCH, self.resize_handler)

            except termios.error:
                tcoldattr = None

            try:
                # TODO: Is socket.SOCK_SEQPACKET supported in Windows?
                with socket.socket(socket.AF_UNIX,
                                   socket.SOCK_SEQPACKET) as skt:
                    # Prepare socket for use with conmon/container
                    skt.connect(self.pseudo_tty.io_socket)

                    sources = [skt, self.pseudo_tty.stdin]
                    while sources:
                        logging.debug('Waiting on sources: %s', sources)
                        readable, _, _ = select.select(sources, [], [])

                        if skt in readable:
                            data = skt.recv(CONMON_BUFSZ)
                            if data:
                                # Remove source marker when writing
                                os.write(self.pseudo_tty.stdout, data[1:])
                            else:
                                sources.remove(skt)

                        if self.pseudo_tty.stdin in readable:
                            data = os.read(self.pseudo_tty.stdin, CONMON_BUFSZ)
                            if data:
                                skt.sendall(data)

                                if self.pseudo_tty.eot in data:
                                    sources.clear()
                            else:
                                sources.remove(self.pseudo_tty.stdin)
            finally:
                if tcoldattr:
                    termios.tcsetattr(self.pseudo_tty.stdin, termios.TCSADRAIN,
                                      tcoldattr)
                    signal.signal(signal.SIGWINCH, signal.SIG_DFL)
            return self._refresh(podman)



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import sys
import unittest

import fixtures
import mock
import testtools
from testtools import content

import tobiko
from tobiko.tests import unit


def canonical_name(cls):
    return __name__ + '.' + cls.__name__


class MyBaseFixture(tobiko.SharedFixture):

    def __init__(self):
        super(MyBaseFixture, self).__init__()
        self.setup_fixture = mock.Mock(
            specs=tobiko.SharedFixture.setup_fixture)
        self.cleanup_fixture = mock.Mock(
            specs=tobiko.SharedFixture.cleanup_fixture)


class MySkyppingFixture(tobiko.SharedFixture):

    def setup_fixture(self):
        tobiko.skip_test('some-reason')

    def cleanup_fixture(self):
        tobiko.skip_test('some-reason')


class MyFixture(MyBaseFixture):
    pass


class GetFixtureTest(unit.TobikoUnitTest):

    def test_by_name(self):
        self._test_get_fixture(canonical_name(MyFixture))

    def test_by_type(self):
        self._test_get_fixture(MyFixture)

    def test_by_instance(self):
        self._test_get_fixture(MyFixture())

    def _test_get_fixture(self, obj):
        fixture = tobiko.get_fixture(obj)
        self.assertIsInstance(fixture, MyFixture)
        self.assertIs(fixture, tobiko.get_fixture(obj))
        if isinstance(obj, fixtures.Fixture):
            self.assertIs(obj, fixture)
        else:
            self.assertIs(fixture, tobiko.get_fixture(
                canonical_name(MyFixture)))
        fixture.setup_fixture.assert_not_called()
        fixture.cleanup_fixture.assert_not_called()

        for fixture_id in range(2):
            other = tobiko.get_fixture(obj, fixture_id=fixture_id)
            if isinstance(obj, fixtures.Fixture) or not fixture_id:
                self.assertIs(fixture, other)
            else:
                self.assertIsNot(fixture, other)


class GetFixtureNameTest(unit.TobikoUnitTest):

    def test_with_instance(self):
        fixture = MyFixture()
        result = tobiko.get_fixture_name(fixture)
        self.assertEqual(canonical_name(MyFixture), result)

    def test_with_other_type(self):
        obj = object()
        ex = self.assertRaises(TypeError, tobiko.get_fixture_name, obj)
        self.assertEqual('Object {obj!r} is not a fixture.'.format(obj=obj),
                         str(ex))


class GetFixtureClassTest(unit.TobikoUnitTest):

    def test_with_name(self):
        result = tobiko.get_fixture_class(canonical_name(MyFixture))
        self.assertIs(MyFixture, result)

    def test_with_type(self):
        result = tobiko.get_fixture_class(MyFixture)
        self.assertIs(MyFixture, result)

    def test_with_instance(self):
        result = tobiko.get_fixture_class(MyFixture())
        self.assertIs(MyFixture, result)


class GetFixtureDirTest(unit.TobikoUnitTest):

    expected_dir = os.path.dirname(__file__)

    def test_with_name(self):
        actual_dir = tobiko.get_fixture_dir(canonical_name(MyFixture))
        self.assertEqual(self.expected_dir, actual_dir)

    def test_with_type(self):
        actual_dir = tobiko.get_fixture_dir(MyFixture)
        self.assertEqual(self.expected_dir, actual_dir)

    def test_with_instance(self):
        actual_dir = tobiko.get_fixture_dir(MyFixture())
        self.assertEqual(self.expected_dir, actual_dir)


class RemoveFixtureTest(unit.TobikoUnitTest):

    def test_with_name(self):
        self._test_remove_fixture(canonical_name(MyFixture))

    def test_with_type(self):
        self._test_remove_fixture(MyFixture)

    def test_with_name_and_fixture_id(self):
        self._test_remove_fixture(canonical_name(MyFixture), fixture_id=5)

    def test_with_type_and_fixture_id(self):
        self._test_remove_fixture(MyFixture, fixture_id=6)

    def _test_remove_fixture(self, obj, fixture_id=None):
        fixture = tobiko.get_fixture(obj, fixture_id=fixture_id)
        result = tobiko.remove_fixture(obj, fixture_id=fixture_id)
        self.assertIs(fixture, result)
        self.assertIsNot(fixture, tobiko.get_fixture(obj,
                                                     fixture_id=fixture_id))
        fixture.setup_fixture.assert_not_called()
        fixture.cleanup_fixture.assert_not_called()


class SetupFixtureTest(unit.TobikoUnitTest):

    def test_with_name(self):
        self._test_setup_fixture(canonical_name(MyFixture))

    def test_with_type(self):
        self._test_setup_fixture(MyFixture)

    def test_with_instance(self):
        self._test_setup_fixture(MyFixture2())

    def test_with_name_and_fixture_id(self):
        self._test_setup_fixture(canonical_name(MyFixture), fixture_id=5)

    def test_with_type_and_fixture_id(self):
        self._test_setup_fixture(MyFixture, fixture_id=6)

    def test_with_instance_and_fixture_id(self):
        self._test_setup_fixture(MyFixture2(), fixture_id=7)

    def _test_setup_fixture(self, obj, fixture_id=None):
        result = tobiko.setup_fixture(obj, fixture_id=fixture_id)
        self.assertIs(tobiko.get_fixture(obj, fixture_id=fixture_id), result)
        result.setup_fixture.assert_called_once_with()
        result.cleanup_fixture.assert_not_called()


class UseFixtureTest(unit.TobikoUnitTest):

    def test_with_name(self):
        self._test_use_fixture(canonical_name(MyFixture))

    def test_with_type(self):
        self._test_use_fixture(MyFixture)

    def test_with_instance(self):
        self._test_use_fixture(MyFixture2())

    def test_with_name_and_fixture_id(self):
        self._test_use_fixture(canonical_name(MyFixture), fixture_id=5)

    def test_with_type_and_fixture_id(self):
        self._test_use_fixture(MyFixture, fixture_id=6)

    def test_with_instance_and_fixture_id(self):
        self._test_use_fixture(MyFixture2(), fixture_id=7)

    def _test_use_fixture(self, obj, fixture_id=None):

        fixture: MyFixture = tobiko.get_fixture(
            obj=obj, fixture_id=fixture_id)  # type: ignore

        class InnerTest(unittest.TestCase):
            def runTest(self):
                fixture.setup_fixture.assert_not_called()
                fixture.cleanup_fixture.assert_not_called()
                result = tobiko.use_fixture(obj, fixture_id=fixture_id)
                fixture.setup_fixture.assert_called_once_with()
                fixture.cleanup_fixture.assert_not_called()
                self.assertIs(fixture, result)

        fixture.setup_fixture.assert_not_called()
        fixture.cleanup_fixture.assert_not_called()

        result = tobiko.run_test(InnerTest())

        fixture.setup_fixture.assert_called_once_with()
        fixture.cleanup_fixture.assert_called_once_with()

        self.assertEqual(1, result.testsRun)
        self.assertEqual([], result.errors)
        self.assertEqual([], result.failures)


class ResetFixtureTest(unit.TobikoUnitTest):

    def test_with_name(self):
        self._test_reset_fixture(canonical_name(MyFixture))

    def test_with_type(self):
        self._test_reset_fixture(MyFixture)

    def test_with_instance(self):
        self._test_reset_fixture(MyFixture2())

    def test_with_name_and_fixture_id(self):
        self._test_reset_fixture(canonical_name(MyFixture), fixture_id=5)

    def test_with_type_and_fixture_id(self):
        self._test_reset_fixture(MyFixture, fixture_id=6)

    def test_with_instance_and_fixture_id(self):
        self._test_reset_fixture(MyFixture(), fixture_id=7)

    def test_after_setup(self):
        fixture = MyFixture2()
        fixture.setUp()
        fixture.setup_fixture.reset_mock()
        self._test_reset_fixture(fixture)

    def test_after_cleanup(self):
        fixture = MyFixture2()
        fixture.cleanUp()
        self._test_reset_fixture(fixture)

    def _test_reset_fixture(self, obj, fixture_id=None, should_clean=True):
        result = tobiko.reset_fixture(obj, fixture_id=fixture_id)
        self.assertIs(tobiko.get_fixture(obj, fixture_id=fixture_id), result)
        result.setup_fixture.assert_called_once_with()
        if should_clean:
            result.cleanup_fixture.assert_called_once_with()
        else:
            result.cleanup_fixture.assert_not_called()


class FailingFixture(tobiko.SharedFixture):

    def setup_fixture(self):
        raise RuntimeError('raised by setup_fixture')

    def cleanup_fixture(self):
        raise RuntimeError('raised by cleanup_fixture')

    def getDetails(self):
        content_object = tobiko.details_content(
            content_type=content.UTF8_TEXT,
            content_id=self.fixture_name,
            get_text=lambda: 'My failure details')
        return {'failing fixture': content_object}


class FailingSetupFixtureWhenFailingTest(unit.TobikoUnitTest):

    def test_with_name(self):
        self._test_setup_fixture(canonical_name(FailingFixture))

    def test_with_type(self):
        self._test_setup_fixture(FailingFixture)

    def test_with_instance(self):
        self._test_setup_fixture(FailingFixture())

    def test_with_name_and_fixture_id(self):
        self._test_setup_fixture(canonical_name(FailingFixture), fixture_id=5)

    def test_with_type_and_fixture_id(self):
        self._test_setup_fixture(FailingFixture, fixture_id=6)

    def test_with_instance_and_fixture_id(self):
        self._test_setup_fixture(FailingFixture(), fixture_id=7)

    def _test_setup_fixture(self, obj, fixture_id=None):
        ex = self.assertRaises(
            RuntimeError, tobiko.setup_fixture, obj, fixture_id=fixture_id)
        self.assertEqual('raised by setup_fixture', str(ex))


class CleanupFixtureTest(unit.TobikoUnitTest):

    def test_with_name(self):
        self._test_cleanup_fixture(canonical_name(MyFixture))

    def test_with_type(self):
        self._test_cleanup_fixture(MyFixture)

    def test_with_instance(self):
        self._test_cleanup_fixture(MyFixture())

    def test_with_name_and_fixture_id(self):
        self._test_cleanup_fixture(canonical_name(MyFixture), fixture_id=5)

    def test_with_type_and_fixture_id(self):
        self._test_cleanup_fixture(MyFixture, fixture_id=6)

    def test_with_instance_and_fixture_id(self):
        self._test_cleanup_fixture(MyFixture(), fixture_id=7)

    def _test_cleanup_fixture(self, obj, fixture_id=None):
        result = tobiko.cleanup_fixture(obj, fixture_id=fixture_id)
        self.assertIs(tobiko.get_fixture(obj, fixture_id=fixture_id), result)
        result.setup_fixture.assert_not_called()
        result.cleanup_fixture.assert_called_once_with()


class MyFixtureWithProperty(MyBaseFixture):

    @tobiko.fixture_property
    def some_property(self):
        return id(self)


class FixturePropertyTest(unit.TobikoUnitTest):

    def test_with_instance(self):
        fixture = tobiko.get_fixture(MyFixtureWithProperty)
        self.assertEqual(id(fixture), fixture.some_property)

    def test_without_instance(self):
        fixture = tobiko.get_fixture(MyFixtureWithProperty)
        self.assertEqual(id(fixture), MyFixtureWithProperty.some_property)


class MyFixture2(MyBaseFixture):
    pass


class MyRequiredFixture(MyBaseFixture):
    pass


class RequiredFixtureTest(unit.TobikoUnitTest):

    required_fixture = tobiko.required_fixture(MyRequiredFixture)
    required_fixture_no_setup = tobiko.required_fixture(
      MyRequiredFixture, setup=False)

    def test_list_required_fixtures_with_module(self):
        module = sys.modules[__name__]
        result = tobiko.list_required_fixtures([module])
        self.assertEqual([], result)

    def test_list_required_fixtures_with_module_name(self):
        result = tobiko.list_required_fixtures([__name__])
        self.assertEqual([], result)

    def test_list_required_fixtures_with_testcase_type(self):
        result = tobiko.list_required_fixtures([RequiredFixtureTest])
        self.assertEqual([canonical_name(MyRequiredFixture)], result)

    def test_list_required_fixtures_with_testcase_name(self):
        result = tobiko.list_required_fixtures(
            [canonical_name(RequiredFixtureTest)])
        self.assertEqual([canonical_name(MyRequiredFixture)], result)

    def test_list_required_fixtures_with_unbound_method(
            self, fixture=MyFixture, fixture2=MyFixture2):
        cls = RequiredFixtureTest
        result = tobiko.list_required_fixtures(
            [cls.test_list_required_fixtures_with_unbound_method])
        self.assertEqual([canonical_name(fixture),
                          canonical_name(fixture2),
                          canonical_name(MyRequiredFixture)], result)

    def test_list_required_fixtures_with_bound_method(
            self, fixture=MyFixture, fixture2=MyFixture2):
        result = tobiko.list_required_fixtures([
            self.test_list_required_fixtures_with_bound_method])
        self.assertEqual([canonical_name(fixture),
                          canonical_name(fixture2),
                          canonical_name(MyRequiredFixture)], result)

    def test_list_required_fixtures_with_method_name(
            self, fixture=MyFixture, fixture2=MyFixture2):
        result = tobiko.list_required_fixtures([self.id()])
        self.assertEqual([canonical_name(fixture),
                          canonical_name(fixture2),
                          canonical_name(MyRequiredFixture)], result)

    def test_list_required_fixtures_with_fixture_name(self):
        result = tobiko.list_required_fixtures([canonical_name(MyFixture)])
        self.assertEqual([canonical_name(MyFixture)], result)

    def test_list_required_fixtures_with_fixture(self):
        result = tobiko.list_required_fixtures([MyFixture()])
        self.assertEqual([canonical_name(MyFixture)], result)

    def test__list_required_fixtures_with_fixture_type(self):
        result = tobiko.list_required_fixtures([MyFixture])
        self.assertEqual([canonical_name(MyFixture)], result)

    def test_required_fixture_with_instance(self):
        fixture = self.required_fixture
        self.assertIsInstance(fixture, MyRequiredFixture)
        fixture.setup_fixture.assert_called()
        fixture.cleanup_fixture.assert_not_called()

    def test_required_fixture_with_no_setup(self):
        fixture = self.required_fixture_no_setup
        self.assertIsInstance(fixture, MyRequiredFixture)
        fixture.setup_fixture.assert_not_called()
        fixture.cleanup_fixture.assert_not_called()


class SharedFixtureTest(unit.TobikoUnitTest):

    def setUp(self):
        super(SharedFixtureTest, self).setUp()
        tobiko.remove_fixture(MyFixture)

    def test_init(self):
        fixture = MyFixture()
        fixture.setup_fixture.assert_not_called()
        fixture.cleanup_fixture.assert_not_called()

    def test_get(self):
        fixture = MyFixture.get()
        self.assertIs(tobiko.get_fixture(MyFixture), fixture)

    def test_use_fixture(self):
        fixture = MyFixture()
        self.addCleanup(fixture.cleanup_fixture.assert_called_once_with)

        self.useFixture(fixture)
        fixture.setup_fixture.assert_called_once_with()
        fixture.cleanup_fixture.assert_not_called()

        self.useFixture(fixture)
        fixture.setup_fixture.assert_called_once_with()
        fixture.cleanup_fixture.assert_not_called()

    def test_add_cleanup(self):
        fixture = MyFixture()
        self.addCleanup(fixture.cleanup_fixture.assert_called_once_with)
        self.addCleanup(fixture.cleanUp)
        self.addCleanup(fixture.cleanUp)

    def test_setup(self):
        fixture = MyFixture()
        fixture.setUp()
        fixture.setup_fixture.assert_called_once_with()

    def test_setup_twice(self):
        fixture = MyFixture()
        fixture.setUp()
        fixture.setUp()
        fixture.setup_fixture.assert_called_once_with()

    def test_setup_when_skipping(self):
        fixture = MySkyppingFixture()
        self.assertRaises(testtools.MultipleExceptions, fixture.setUp)
        self.assertRaises(testtools.MultipleExceptions, fixture.setUp)

    def test_cleanup(self):
        fixture = MyFixture()
        fixture.cleanUp()
        fixture.cleanup_fixture.assert_called_once_with()

    def test_cleanup_twice(self):
        fixture = MyFixture()
        fixture.cleanUp()
        fixture.cleanUp()
        fixture.cleanup_fixture.assert_called_once_with()

    def test_cleanup_when_skipping(self):
        fixture = MySkyppingFixture()
        self.assertRaises(tobiko.SkipException, fixture.cleanUp)
        self.assertRaises(testtools.MultipleExceptions, fixture.cleanUp)

    def test_lifecycle(self):
        fixture = MyFixture()

        for call_count in range(3):
            fixture.setUp()
            fixture.setup_fixture.assert_has_calls([mock.call()] * call_count)
            fixture.setUp()
            fixture.setup_fixture.assert_has_calls([mock.call()] * call_count)

            fixture.cleanUp()
            fixture.cleanup_fixture.assert_has_calls(
                [mock.call()] * call_count)
            fixture.cleanUp()
            fixture.cleanup_fixture.assert_has_calls(
                [mock.call()] * call_count)

    def test_fixture_name(self):
        fixture = tobiko.get_fixture(MyFixture)
        self.assertEqual(f'{__name__}.{MyFixture.__qualname__}',
                         fixture.fixture_name)

    def test_fixture_name_with_fixture_id(self):
        fixture = tobiko.get_fixture(MyFixture, fixture_id=10)
        self.assertEqual(f'{__name__}.{MyFixture.__qualname__}-10',
                         fixture.fixture_name)

    def test_fixture_id(self):
        fixture = tobiko.get_fixture(MyFixture)
        self.assertIsNone(fixture.fixture_id)

    def test_fixture_id_with_fixture_id(self):
        fixture = tobiko.get_fixture(MyFixture, fixture_id=12)
        self.assertEqual(12, fixture.fixture_id)



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import ast
import datetime
import re
import typing

from oslo_log import log

import tobiko
from tobiko.openstack import neutron
from tobiko.openstack.neutron import _agent
from tobiko.openstack import topology
from tobiko import podified
from tobiko.shell import files


LOG = log.getLogger(__name__)


class NeutronNovaResponse(typing.NamedTuple):
    hostname: str
    timestamp: float
    name: str
    server_uuid: str
    status: str
    code: int
    tag: str
    line: str

    def __lt__(self, other):
        return self.timestamp < other.timestamp


class NeutronNovaCommonReader(tobiko.SharedFixture):
    log_digger: files.MultihostLogFileDigger
    groups: typing.List[str]
    message_pattern: str
    datetime_pattern: typing.Pattern
    config = tobiko.required_fixture(topology.OpenStackTopologyConfig)
    service_name = neutron.SERVER

    def setup_fixture(self):
        self.datetime_pattern = re.compile(
            self.config.conf.log_datetime_pattern)
        self.log_digger = self.useFixture(
            topology.get_log_file_digger(
                service_name=self.service_name,
                groups=self.groups,
                pattern=self.message_pattern))
        self.read_responses()

    def _get_log_timestamp(self,
                           log_line: str) -> float:
        found = self.datetime_pattern.match(log_line)
        if not found:
            return 0.0
        return datetime.datetime.strptime(
            found.group(1), "%Y-%m-%d %H:%M:%S.%f").timestamp()

    def read_responses(self):
        raise NotImplementedError


class NeutronNovaResponseReader(NeutronNovaCommonReader):
    groups = ['controller']
    message_pattern = r'Nova event response: '
    responses: tobiko.Selection[NeutronNovaResponse]

    def read_responses(self) -> tobiko.Selection[NeutronNovaResponse]:
        # pylint: disable=no-member
        responses = tobiko.Selection[NeutronNovaResponse]()
        message_pattern = re.compile(self.message_pattern)
        for hostname, line in self.log_digger.find_lines(
                new_lines=hasattr(self, 'responses')):
            found = message_pattern.search(line)
            assert found is not None
            response_text = line[found.end():].strip()
            response_data = ast.literal_eval(response_text)
            assert isinstance(response_data, dict)
            response = NeutronNovaResponse(
                hostname=hostname,
                line=line,
                timestamp=self._get_log_timestamp(line[:found.start()]),
                **response_data)
            responses.append(response)
        responses.sort()
        if hasattr(self, 'responses'):
            self.responses.extend(responses)
        else:
            self.responses = responses
        return responses


def read_neutron_nova_responses(
        reader: NeutronNovaResponseReader = None,
        new_lines=True,
        **attributes) \
        -> tobiko.Selection[NeutronNovaResponse]:
    if reader is None:
        reader = tobiko.setup_fixture(NeutronNovaResponseReader)
    responses = reader.read_responses()
    if not new_lines:
        responses = reader.responses
    if attributes and responses:
        responses = responses.with_attributes(**attributes)
    return responses


class UnsupportedDhcpOptionMessage(typing.NamedTuple):
    port_uuid: str
    unsupported_dhcp_option: str
    timestamp: float
    line: str

    def __lt__(self, other):
        return self.timestamp < other.timestamp


@podified.skip_if_podified
@_agent.skip_unless_is_ovn()
class OvnUnsupportedDhcpOptionReader(NeutronNovaCommonReader):
    groups = ['controller']
    message_pattern = (
        'The DHCP option .* on port .* is not suppported by OVN, ignoring it')
    responses: tobiko.Selection[UnsupportedDhcpOptionMessage]

    def read_responses(self) \
            -> tobiko.Selection[UnsupportedDhcpOptionMessage]:
        # pylint: disable=no-member
        def _get_port_uuid(line):
            port_pattern = 'on port (.*) is not suppported by OVN'
            return re.findall(port_pattern, line)[0]

        def _get_dhcp_option(line):
            dhcp_opt_pattern = 'The DHCP option (.*) on port'
            return re.findall(dhcp_opt_pattern, line)[0]

        responses = tobiko.Selection[UnsupportedDhcpOptionMessage]()
        message_pattern = re.compile(self.message_pattern)
        for _, line in self.log_digger.find_lines(
                new_lines=hasattr(self, 'responses')):
            found = message_pattern.search(line)
            assert found is not None
            response = UnsupportedDhcpOptionMessage(
                line=line,
                timestamp=self._get_log_timestamp(line[:found.start()]),
                port_uuid=_get_port_uuid(line),
                unsupported_dhcp_option=_get_dhcp_option(line))
            responses.append(response)
        responses.sort()
        if hasattr(self, 'responses'):
            self.responses.extend(responses)
        else:
            self.responses = responses
        return responses


def assert_ovn_unsupported_dhcp_option_messages(
        reader: OvnUnsupportedDhcpOptionReader = None,
        new_lines=True,
        unsupported_options: typing.Optional[typing.List] = None,
        **attributes):
    if reader is None:
        reader = tobiko.setup_fixture(OvnUnsupportedDhcpOptionReader)
    # find new logs that match the pattern
    responses = reader.read_responses()
    if not new_lines:
        responses = reader.responses
    if attributes and responses:
        responses = responses.with_attributes(**attributes)

    # assert one line matches per unsupported dhcp option
    test_case = tobiko.get_test_case()
    for unsupported_option in unsupported_options or []:
        messages_unsupported_option = responses.with_attributes(
            unsupported_dhcp_option=unsupported_option)
        test_case.assertEqual(1, len(messages_unsupported_option))
        LOG.debug('Found one match for unsupported dhcp option '
                  f'{unsupported_option}')



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr

import tobiko
from tobiko.openstack import nova
from tobiko.openstack import neutron
from tobiko.shiftstack import _keystone


def shiftstack_neutron_client(obj: neutron.NeutronClientType) \
        -> neutron.NeutronClient:
    if obj is None:
        return get_shiftstack_neutron_client()
    else:
        return tobiko.check_valid_type(obj, neutron.NeutronClient)


def get_shiftstack_neutron_client() -> neutron.NeutronClient:
    session = _keystone.shiftstack_keystone_session()
    return neutron.get_neutron_client(session=session)


def list_shiftstack_node_ip_addresses(
        server: nova.ServerType,
        ip_version: int = None,
        client: neutron.NeutronClientType = None) \
        -> tobiko.Selection[netaddr.IPAddress]:
    client = shiftstack_neutron_client(client)
    return neutron.list_device_ip_addresses(device=server,
                                            ip_version=ip_version,
                                            client=client)


def find_shiftstack_node_ip_address(
        server: nova.ServerType,
        ip_version: int = None,
        client: neutron.NeutronClientType = None,
        unique=False) -> netaddr.IPAddress:
    addresses = list_shiftstack_node_ip_addresses(server=server,
                                                  ip_version=ip_version,
                                                  client=client)
    if unique:
        return addresses.unique
    else:
        return addresses.first



# Copyright (c) 2023 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
from oslo_log import log
import dns.resolver

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import topology
from tobiko.openstack import stacks
from tobiko.shell import ip

LOG = log.getLogger(__name__)


@keystone.skip_if_missing_service(name='designate')
class DesignateBasicScenarioTest(testtools.TestCase):
    """Designate sanity scenario test.
    In this test we'll create a Zone and A type recordset
    associated to that zone, then we'll check: that SOA,
     NS and A recordsets are listed in "openstack recordset list"
    and we'll also check (using Dig) that the BIND9 server has
     been updated with the same records: SOA, NS and A.
    """

    zone_stack = tobiko.required_fixture(stacks.DesignateZoneStackFixture)

    def setUp(self):
        super(DesignateBasicScenarioTest, self).setUp()
        # Wait for Designate recourses to be active
        self.zone_stack.wait_for_active_recordsets()

    def test_basic_sanity_scenario(self):

        # check that SOA, NS and A recordsets are listed in recordset list
        recordset_list = self.zone_stack.recordset_list
        expected_recordset_types = ["SOA", "NS", "A"]
        actual_recordset_types = [item["type"] for item in recordset_list]
        expected_recordset_types.sort()
        actual_recordset_types.sort()
        self.assertEqual(expected_recordset_types, actual_recordset_types)

        # list all Controller nodes
        nodes = topology.list_openstack_nodes(group='controller')
        valid_dns_ip_list = []

        for node in nodes:
            LOG.debug(node.ssh_parameters.get('hostname'))

            my_resolver = dns.resolver.Resolver()

            controller_ip_list = ip.list_ip_addresses(
                ssh_client=node.ssh_client, ip_version=4, scope='global')
            for dns_ip in controller_ip_list:
                my_resolver.nameservers = [str(dns_ip)]
                try:
                    answerSOA = my_resolver.resolve(
                        'tobiko.openstack.stacks._designate'
                        '.designatezonestackfixture.', 'SOA')
                    answerNS = my_resolver.resolve(
                        'tobiko.openstack.stacks.'
                        '_designate.designatezonestackfixture.', 'NS')
                    answerA = my_resolver.resolve(
                        'record.tobiko.openstack.stacks._designate.'
                        'designatezonestackfixture.', 'A')
                    LOG.debug(f"{answerSOA},{answerNS},{answerA}"
                              f" is stored for a zone")
                    LOG.debug(f"{dns_ip} is a valid DNS server!")
                    valid_dns_ip_list.append(str(dns_ip))
                except (dns.resolver.Timeout, dns.resolver.NXDOMAIN):
                    LOG.warn(f"{dns_ip} is NOT a valid DNS server!")

        self.assertEqual(len(valid_dns_ip_list), len(nodes))
        LOG.debug(f"List of valid BIND9 IPs: {valid_dns_ip_list}")



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.shell import sh
from tobiko import tripleo


LOG = log.getLogger(__name__)


@keystone.skip_if_missing_service(name='octavia')
@tripleo.skip_if_missing_overcloud
class OctaviaBasicFaultTest(testtools.TestCase):
    """Octavia fault scenario test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balancer VIP.
    Restart the amphora's compute node to create a failover.
    Reach the members to make sure they are ready to be checked.
    Generate network traffic again to verify Octavia functionality.
    """
    lb = None
    listener = None
    pool = None
    server_stack = tobiko.required_fixture(
        stacks.UbuntuServerStackFixture)
    other_server_stack = tobiko.required_fixture(
        stacks.OctaviaOtherServerStackFixture)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaBasicFaultTest, self).setUp()

        self.lb, self.listener, self.pool = octavia.deploy_ipv4_amphora_lb(
            servers_stacks=[self.server_stack, self.other_server_stack]
        )

        self._send_http_traffic()

    def _send_http_traffic(self):
        # For 5 minutes we ignore specific exceptions as we know
        # that Octavia resources are being provisioned
        for attempt in tobiko.retry(timeout=300.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.pool.id,
                    ip_address=self.lb.vip_address,
                    lb_algorithm=self.pool.lb_algorithm,
                    protocol=self.listener.protocol,
                    port=self.listener.protocol_port)
                break
            except (octavia.RoundRobinException,
                    octavia.TrafficTimeoutError,
                    sh.ShellCommandFailed):
                LOG.exception(
                    f"Traffic didn't reach all members after "
                    f"#{attempt.number} attempts and "
                    f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise

    def test_reboot_amphora_compute_node(self):
        amphora_compute_host = octavia.get_amphora_compute_node(
            load_balancer_id=self.lb.id,
            port=self.listener.protocol_port,
            protocol=self.listener.protocol,
            ip_address=self.lb.vip_address)

        LOG.debug('Rebooting compute node...')

        # Reboot Amphora's compute node will initiate a failover
        amphora_compute_host.reboot_node()

        LOG.debug('Compute node has been rebooted')

        # Wait for the LB to be updated
        try:
            octavia.wait_for_status(object_id=self.lb.id)

        # The reboot_overcloud_node function restarts other running nova
        # vms/backend servers after the compute node reboot by default.
        # Those restart operations may take longer than the LB transitions into
        # PENDING_UPDATE and then into ACTIVE altogether. So if the restarted
        # vms will finish their restart process after the LB reaches ACTIVE
        # status, the lb will never reach PENDING_UPDATE
        except tobiko.RetryTimeLimitError:
            LOG.info('The restarted servers/backend members reached ACTIVE '
                     'status after the LB finished its update process, hence '
                     'no exception is being raised even though the update '
                     'timeout was reached.')

        octavia.wait_for_status(object_id=self.lb.id)

        LOG.debug(f'Load Balancer {self.lb.id} is ACTIVE')

        self._send_http_traffic()

    def test_kill_amphora_agent(self):
        """Kill the MASTER amphora agent

        This test kills the amphora agent on the MASTER amphora.
        Killing the amphora agent will cause a failover.

        Octavia's functionality will be verified afterwards.
        """

        self._skip_if_not_active_standby()

        # Finding the amphora agent pid and kill it
        amp_agent_pid_command = (
            "ps -ef | awk '/amphora/{print $2}' | head -n 1")
        amp_agent_pid = octavia.run_command_on_amphora(
            command=amp_agent_pid_command,
            lb_id=self.lb.id,
            lb_vip=self.lb.vip_address)
        LOG.info(f'The amp_agent_pid is {amp_agent_pid}')

        octavia.run_command_on_amphora(
            command=f'kill -9 {amp_agent_pid}',
            lb_id=self.lb.id,
            lb_vip=self.lb.vip_address,
            sudo=True)

        self._wait_for_failover_and_test_functionality()

    def test_stop_keepalived(self):
        """Stop keepalived on MASTER amphora

        This test stops keepalived on the MASTER amphora.
        Stopping keepalived on the amphora will cause a failover.

        Octavia's functionality will be verified afterwards.
        """

        self._skip_if_not_active_standby()

        stop_keepalived_cmd = 'systemctl stop octavia-keepalived'

        octavia.run_command_on_amphora(
            command=stop_keepalived_cmd,
            lb_id=self.lb.id,
            lb_vip=self.lb.vip_address,
            sudo=True)

        self._wait_for_failover_and_test_functionality()

    def test_stop_haproxy(self):
        """Stop haproxy on MASTER amphora

        This test stops haproxy on the MASTER amphora.
        Stopping haproxy on the amphora will cause a failover.

        Octavia's functionality will be verified afterwards.
        """

        self._skip_if_not_active_standby()

        # Finding the amphora haproxy unit name and stop it
        amp_haproxy_unit_command = (
            "systemctl list-units | awk '/haproxy-/{print $1}'")

        for attempt in tobiko.retry(timeout=300., interval=5.):
            amp_haproxy_unit = octavia.run_command_on_amphora(
                command=amp_haproxy_unit_command,
                lb_id=self.lb.id,
                lb_vip=self.lb.vip_address)

            if amp_haproxy_unit:
                break

            # it will raise tobiko.RetryTimeLimitError in case of timeout
            attempt.check_limits()

            LOG.debug(f"Waiting for haproxy instance on LB {self.lb.id} to get"
                      " the haproxy unit id")

        LOG.info(f'The amp_haproxy_unit is {amp_haproxy_unit}')

        octavia.run_command_on_amphora(
            command=f'systemctl stop {amp_haproxy_unit}',
            lb_id=self.lb.id,
            lb_vip=self.lb.vip_address,
            sudo=True)

        self._wait_for_failover_and_test_functionality()

    def _skip_if_not_active_standby(self):
        """Skip the test if Octavia doesn't use Active/standby topology
        """
        if len(list(octavia.list_amphorae(load_balancer_id=self.lb.id))) < 2:
            skipping_stmt = ('Skipping the test as it requires '
                             'Active/standby topology.')
            LOG.info(skipping_stmt)
            self.skipTest(skipping_stmt)

    def _wait_for_failover_and_test_functionality(self, timeout=70):
        """Wait for failover to end and test Octavia functionality"""

        octavia.wait_for_status(object_id=self.lb.id,
                                status=octavia.PENDING_UPDATE,
                                timeout=timeout)

        octavia.wait_for_status(object_id=self.lb.id)

        LOG.debug(f'Load Balancer {self.lb.id} is ACTIVE')

        self._send_http_traffic()



from __future__ import absolute_import

import abc
import os
import re
import typing

from oslo_log import log

import tobiko
from tobiko.openstack import topology
from tobiko import podman
from tobiko import docker
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


expected_containers_file = os.path.expanduser(
    '~/expected_containers_list_df.csv')


class ContainerRuntime(abc.ABC):
    runtime_name: str
    version_pattern: typing.Pattern

    def match_version(self, version: str) -> bool:
        for version_line in version.splitlines():
            if self.version_pattern.match(version_line) is not None:
                return True
        return False

    def get_client(self, ssh_client):
        for attempt in tobiko.retry(timeout=60.0,
                                    interval=5.0):
            try:
                client = self._get_client(ssh_client=ssh_client)
                break
            # TODO chose a better exception type
            except Exception:
                if attempt.is_last:
                    raise
                LOG.debug('Unable to connect to docker server',
                          exc_info=1)
                ssh.reset_default_ssh_port_forward_manager()
        else:
            raise RuntimeError("Broken retry loop")
        return client

    def _get_client(self, ssh_client):
        raise NotImplementedError

    def list_containers(self, ssh_client):
        raise NotImplementedError


class DockerContainerRuntime(ContainerRuntime):
    runtime_name = 'docker'
    version_pattern = re.compile('Docker version .*', re.IGNORECASE)

    def _get_client(self, ssh_client):
        return docker.get_docker_client(ssh_client=ssh_client,
                                        sudo=True).connect()

    def list_containers(self, ssh_client):
        client = self.get_client(ssh_client=ssh_client)
        return docker.list_docker_containers(client=client)


class PodmanContainerRuntime(ContainerRuntime):
    runtime_name = 'podman'
    version_pattern = re.compile('Podman version .*', re.IGNORECASE)

    def _get_client(self, ssh_client):
        return podman.get_podman_client(ssh_client=ssh_client).connect()

    def list_containers(self, ssh_client):
        client = self.get_client(ssh_client=ssh_client)
        return podman.list_podman_containers(client=client)


DOCKER_RUNTIME = DockerContainerRuntime()
PODMAN_RUNTIME = PodmanContainerRuntime()
CONTAINER_RUNTIMES = [PODMAN_RUNTIME, DOCKER_RUNTIME]


class ContainerMismatchException(tobiko.TobikoException):
    pass


def remove_containers_from_comparison(comparable_containers_df):
    """remove any containers if comparing them with previous status is not
    necessary or makes no sense
    """
    os_topology = topology.get_openstack_topology()
    for row in comparable_containers_df.iterrows():
        for ignore_container in os_topology.ignore_containers_list:
            if ignore_container in str(row):
                LOG.info(f'container {ignore_container} has changed state, '
                         'but that\'s ok - it will be ignored and the test '
                         f'will not fail due to this: {str(row)}')
                # if a pcs resource is found , we drop that row
                comparable_containers_df.drop(row[0], inplace=True)
                # this row was already dropped, go to next row
                break


def dataframe_difference(df1, df2, which=None):
    """Find rows which are different between two DataFrames."""
    comparison_df = df1.merge(df2,
                              indicator='same_state',
                              how='outer')
    # return only non identical rows
    if which is None:
        diff_df = comparison_df[comparison_df['same_state'] != 'both']

    else:
        diff_df = comparison_df[comparison_df['same_state'] == which]

    # if the list of different state containers includes sidecar containers,
    # ignore them because the existence of these containers depends on the
    # created resources
    # if the list of different state containers includes pacemaker resources,
    # ignore them since the sanity and fault tests check pacemaker status too
    remove_containers_from_comparison(diff_df)

    return diff_df



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc
import asyncio
import typing


class ActorRequest(typing.NamedTuple):
    future: asyncio.Future
    actor_id: str
    method: str
    arguments: typing.Dict[str, typing.Any]


class ActorRequestQueue(abc.ABC):

    @abc.abstractmethod
    def send_request(self,
                     actor_id: str,
                     method: str,
                     arguments: typing.Dict[str, typing.Any]) \
            -> asyncio.Future:
        raise NotImplementedError

    async def cancel_requests(self,
                              actor_id: str = None) \
            -> typing.List[ActorRequest]:
        return await self.drain_requests(actor_id=actor_id,
                                         cancel=True)

    @abc.abstractmethod
    async def drain_requests(self,
                             actor_id: str = None,
                             cancel=False) -> typing.List[ActorRequest]:
        raise NotImplementedError

    @abc.abstractmethod
    async def receive_request(self) -> ActorRequest:
        raise NotImplementedError


class AsyncioActorRequestQueue(ActorRequestQueue):

    def __init__(self,
                 loop: asyncio.AbstractEventLoop,
                 max_size=0):
        self.max_size = max_size
        self._loop = loop
        self._queue: asyncio.Queue = self._init_queue()

    def _init_queue(self) -> asyncio.Queue:
        return asyncio.Queue(maxsize=self.max_size)

    def send_request(self,
                     actor_id: str,
                     method: str,
                     arguments: typing.Dict[str, typing.Any]) \
            -> asyncio.Future:
        future = self._loop.create_future()
        request = ActorRequest(future=future,
                               actor_id=actor_id,
                               method=method,
                               arguments=arguments)
        self._queue.put_nowait(request)
        return future

    async def drain_requests(self, actor_id: str = None,
                             cancel=False) \
            -> typing.List[ActorRequest]:
        old_queue = self._queue
        self._queue = self._init_queue()
        keep_requests = []
        drained_requests = []
        while True:
            try:
                request: ActorRequest = old_queue.get_nowait()
            except asyncio.QueueEmpty:
                break
            if actor_id in [None, request.actor_id]:
                drained_requests.append(request)
                if cancel:
                    request.future.cancel()
            else:
                keep_requests.append(request)
        for request in keep_requests:
            await self._queue.put(request)
        return drained_requests

    async def receive_request(self) -> ActorRequest:
        return await self._queue.get()


def create_request_queue(loop: asyncio.AbstractEventLoop,
                         max_size=0) -> ActorRequestQueue:
    return AsyncioActorRequestQueue(loop=loop, max_size=max_size)



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import typing

from oslo_log import log

import tobiko
from tobiko.openstack import octavia
from tobiko.shell import curl
from tobiko.shell import ssh
from tobiko.shell import sh


LOG = log.getLogger(__name__)


def check_members_balanced(ip_address: str,
                           protocol: str,
                           port: int,
                           pool_id: str = None,
                           members_count: int = None,
                           lb_algorithm: str = None,
                           requests_count: int = 10,
                           connect_timeout: tobiko.Seconds = 10.,
                           interval: tobiko.Seconds = 1,
                           ssh_client: ssh.SSHClientFixture = None) -> (
        typing.Dict[str, int]):

    """Check if traffic is properly balanced between members."""

    # Getting the members count
    if members_count is None:
        if pool_id is None:
            raise ValueError('Either members_count or pool_id has to be passed'
                             ' to the function.')

        else:  # members_count is None and pool_id is not None
            members_count = len(list(octavia.list_members(pool_id=pool_id)))

    last_content = None
    replies: typing.Dict[str, int] = collections.defaultdict(lambda: 0)
    for attempt in tobiko.retry(count=members_count * requests_count,
                                interval=interval):
        try:
            content = curl.execute_curl(
                hostname=ip_address,
                scheme='HTTP' if protocol == 'TCP' else protocol,
                port=port,
                path='id',
                connect_timeout=connect_timeout,
                ssh_client=ssh_client).strip()
        except sh.ShellCommandFailed as ex:
            if ex.exit_status == 28:
                raise octavia.TrafficTimeoutError(
                    reason=str(ex.stderr)) from ex
            else:
                raise ex

        replies[content] += 1

        if last_content is not None and lb_algorithm == 'ROUND_ROBIN':
            if members_count > 1 and last_content == content:
                raise octavia.RoundRobinException(
                    'Request was forwarded two times to the same host:\n'
                    f'members_count: {members_count}\n'
                    f'expected: {last_content}\n'
                    f'actual: {content}\n')

        last_content = content

        if attempt.is_last:
            break
    else:
        raise RuntimeError('Broken retry loop')

    LOG.debug(f"Replies counts from load balancer: {replies}")

    # assert that 'members_count' servers replied
    missing_members_count = members_count - len(replies)
    LOG.debug(f'Members count from pool {pool_id} is {members_count}')
    LOG.debug(f'len(replies) is {len(replies)}')

    if 0 != missing_members_count:
        raise octavia.RoundRobinException(
            f'Missing replies from {missing_members_count} members.')

    return replies



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import re
import typing

import tobiko
from tobiko.tests import unit


def condition(value):
    return value


class Obj(typing.NamedTuple):
    number: int = 0
    text: str = ''


T = typing.TypeVar('T')


class SelectionTest(unit.TobikoUnitTest):

    @staticmethod
    def create_selection(*args, **kwargs):
        return tobiko.Selection(*args, **kwargs)

    def test_selection(self,
                       objects: typing.Iterable[T] = tuple()) \
            -> tobiko.Selection[T]:
        reference = list(objects)
        if isinstance(objects, abc.Generator):
            # Can't reiterate the same generator twice
            objects = (o for o in reference)
            assert isinstance(objects, abc.Generator)
        elif isinstance(objects, abc.Iterator):
            # Can't reiterate the same iterator twice
            objects = iter(reference)
            assert isinstance(objects, abc.Iterator)

        selection = self.create_selection(objects)
        self.assertIsInstance(selection, list)
        self.assertIsInstance(selection, tobiko.Selection)
        self.assertEqual(reference, selection)
        self.assertEqual(selection, reference)
        return selection

    def test_selection_with_list(self):
        self.test_selection([1, 'a', 3.14])

    def test_selection_with_tuple(self):
        self.test_selection((1, 'a', 3.14))

    def test_selection_with_generator(self):
        self.test_selection(c for c in 'hello')

    def test_selection_with_iterator(self):
        self.test_selection(iter('hello'))

    def test_with_attribute(self):
        a = Obj(0, 'a')
        b = Obj(1, 'b')
        c = Obj(1, 'c')
        selection = self.create_selection([a, b, c])
        self.assertEqual([a], selection.with_attributes(text='a'))
        self.assertEqual([b], selection.with_attributes(text='b'))
        self.assertEqual([c], selection.with_attributes(text='c'))
        self.assertEqual([a], selection.with_attributes(number=0))
        self.assertEqual([b, c], selection.with_attributes(number=1))
        self.assertEqual([], selection.with_attributes(number=2))
        self.assertEqual([b], selection.with_attributes(number=1, text='b'))
        self.assertEqual([], selection.with_attributes(number=1, text='a'))
        self.assertEqual([a, b],
                         selection.with_attributes(text=re.compile('(a|b)')))

    def test_without_attribute(self):
        a = Obj(0, 'a')
        b = Obj(1, 'b')
        c = Obj(1, 'c')
        selection = self.create_selection([a, b, c])
        self.assertEqual([b, c], selection.without_attributes(text='a'))
        self.assertEqual([a, c], selection.without_attributes(text='b'))
        self.assertEqual([a, b], selection.without_attributes(text='c'))
        self.assertEqual([b, c], selection.without_attributes(number=0))
        self.assertEqual([a], selection.without_attributes(number=1))
        self.assertEqual([a, b, c], selection.without_attributes(number=2))
        self.assertEqual([a], selection.without_attributes(number=1, text='b'))
        self.assertEqual([], selection.without_attributes(number=1, text='a'))
        self.assertEqual([c],
                         selection.without_attributes(
                             text=re.compile('(a|b)')))

    def test_with_items(self):
        a = {'number': 0, 'text': 'a'}
        b = {'number': 1, 'text': 'b'}
        c = {'number': 1, 'text': 'c'}
        selection = self.create_selection([a, b, c])
        self.assertEqual([a], selection.with_items(text='a'))
        self.assertEqual([b], selection.with_items(text='b'))
        self.assertEqual([c], selection.with_items(text='c'))
        self.assertEqual([a], selection.with_items(number=0))
        self.assertEqual([b, c], selection.with_items(number=1))
        self.assertEqual([], selection.with_items(number=2))
        self.assertEqual([b], selection.with_items(number=1, text='b'))
        self.assertEqual([], selection.with_items(number=1, text='a'))
        self.assertEqual([a, b],
                         selection.with_items(text=re.compile('(a|b)')))

    def test_without_items(self):
        a = {'number': 0, 'text': 'a'}
        b = {'number': 1, 'text': 'b'}
        c = {'number': 1, 'text': 'c'}
        selection = self.create_selection([a, b, c])
        self.assertEqual([b, c], selection.without_items(text='a'))
        self.assertEqual([a, c], selection.without_items(text='b'))
        self.assertEqual([a, b], selection.without_items(text='c'))
        self.assertEqual([b, c], selection.without_items(number=0))
        self.assertEqual([a], selection.without_items(number=1))
        self.assertEqual([a, b, c], selection.without_items(number=2))
        self.assertEqual([a], selection.without_items(number=1, text='b'))
        self.assertEqual([], selection.without_items(number=1, text='a'))
        self.assertEqual([c],
                         selection.without_items(text=re.compile('(a|b)')))

    def test_select(self):
        a = Obj(0, 'a')
        b = Obj(1, 'b')
        c = Obj(1, 'c')
        selection = self.create_selection([a, b, c])
        self.assertEqual([b, c], selection.select(lambda obj: obj.number == 1))

    def test_unselect(self):
        a = Obj(0, 'a')
        b = Obj(1, 'b')
        c = Obj(1, 'c')
        selection = self.create_selection([a, b, c])
        self.assertEqual([a], selection.unselect(lambda obj: obj.number == 1))

    def test_select_uniques(self):
        a = Obj(0, 'a')
        b = Obj(1, 'b')
        c = Obj(1, 'c')
        selection = self.create_selection([a, b, c, a, b, c])
        self.assertEqual([a, b, c], selection.select_uniques())


class SelectTest(SelectionTest):

    @staticmethod
    def create_selection(*args, **kwargs):
        return tobiko.select(*args, **kwargs)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import metalsmith
import netaddr

import tobiko
from tobiko.shell import ping
from tobiko.shell import ssh
from tobiko.openstack.metalsmith import _client


MetalsmithInstance = typing.Union[metalsmith.Instance]


def list_instances(client: _client.MetalsmithClientType = None,
                   **params) \
        -> tobiko.Selection[MetalsmithInstance]:
    instances = tobiko.select(
        _client.metalsmith_client(client).list_instances())
    if params:
        instances = instances.with_attributes(**params)
    return instances


def find_instance(client: _client.MetalsmithClientType = None,
                  unique=False, **params) -> MetalsmithInstance:
    servers = list_instances(client=client, **params)
    if unique:
        return servers.unique
    else:
        return servers.first


def list_instance_ip_addresses(instance: MetalsmithInstance,
                               ip_version: int = None,
                               network_name: str = None,
                               check_connectivity=False,
                               ssh_client: ssh.SSHClientType = None)\
        -> tobiko.Selection[netaddr.IPAddress]:
    ip_addresses = tobiko.Selection[netaddr.IPAddress]()
    for _network, addresses in instance.ip_addresses().items():
        if network_name not in [None, _network]:
            continue

        for address in addresses:
            ip_address = netaddr.IPAddress(address)
            if ip_version not in [None, ip_address.version]:
                continue
            ip_addresses.append(ip_address)

    # check ICMP connectivity
    if check_connectivity:
        ip_addresses = ping.list_reachable_hosts(
            ip_addresses, ssh_client=ssh_client)

    return ip_addresses


def find_instance_ip_address(instance: MetalsmithInstance,
                             unique=False,
                             **params) -> netaddr.IPAddress:
    addresses = list_instance_ip_addresses(instance=instance, **params)
    if unique:
        return addresses.unique
    else:
        return addresses.first



# Copyright 2023 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.shell import sh

CONF = config.CONF
LOG = log.getLogger(__name__)

OSP_CONTROLPLANE = 'openstackcontrolplane'
OSP_DP_NODESET = 'openstackdataplanenodeset'
DP_SSH_SECRET_NAME = 'secret/dataplane-ansible-ssh-private-key-secret'
OSP_BM_HOST = 'baremetalhost.metal3.io'
OSP_BM_CRD = 'baremetalhosts.metal3.io'
OCP_WORKERS = 'nodes'
OVNDBCLUSTER = 'ovndbcluster'

OVN_DP_SERVICE_NAME = 'ovn'
COMPUTE_DP_SERVICE_NAMES = ['nova', 'nova-custom', 'nova-custom-ceph']

EDPM_COMPUTE_GROUP = 'edpm-compute'
EDPM_NETWORKER_GROUP = 'edpm-networker'
EDPM_OTHER_GROUP = 'edpm-other'


_IS_OC_CLIENT_AVAILABLE = None
_IS_BM_CRD_AVAILABLE = None

try:
    import openshift_client as oc
except ModuleNotFoundError:
    _IS_OC_CLIENT_AVAILABLE = False


def _is_oc_client_available() -> bool:
    # pylint: disable=global-statement
    global _IS_OC_CLIENT_AVAILABLE
    if _IS_OC_CLIENT_AVAILABLE is None:
        _IS_OC_CLIENT_AVAILABLE = False
        try:
            if sh.execute('which oc').exit_status == 0:
                _IS_OC_CLIENT_AVAILABLE = True
        except sh.ShellCommandFailed:
            pass
    return _IS_OC_CLIENT_AVAILABLE


def _is_baremetal_crd_available() -> bool:
    # pylint: disable=global-statement
    global _IS_BM_CRD_AVAILABLE
    if not _is_oc_client_available():
        return False
    if _IS_BM_CRD_AVAILABLE is None:
        try:
            # oc.selector("crd") does not need to run on a specific OCP project
            _IS_BM_CRD_AVAILABLE = any(
                [OSP_BM_CRD in n for n in oc.selector("crd").qnames()])
        except oc.OpenShiftPythonException:
            _IS_BM_CRD_AVAILABLE = False
    return _IS_BM_CRD_AVAILABLE


def _get_group(services):
    for compute_dp_service in COMPUTE_DP_SERVICE_NAMES:
        if compute_dp_service in services:
            return EDPM_COMPUTE_GROUP
    if OVN_DP_SERVICE_NAME in services:
        return EDPM_NETWORKER_GROUP
    return EDPM_OTHER_GROUP


def _get_ocp_worker_hostname(worker):
    for address in worker.get('status', {}).get('addresses', []):
        if address.get('type') == 'Hostname':
            return address['address']


def _get_ocp_worker_addresses(worker):
    return [
        netaddr.IPAddress(address['address']) for
        address in worker.get('status', {}).get('addresses', [])
        if address.get('type') != 'Hostname']


def _get_edpm_node_ctlplane_ip_from_status(hostname, node_status):
    all_ips = node_status.get('AllIPs') or node_status.get('allIPs')
    if not all_ips:
        LOG.warning("No IPs found in the Nodeset status: %s",
                    node_status)
        return
    host_ips = all_ips.get(hostname)
    if not host_ips:
        LOG.warning("Host %s not found in AllIPs: %s",
                    hostname, all_ips)
        return
    return host_ips.get('ctlplane')


def has_podified_cp() -> bool:
    if not _is_oc_client_available():
        LOG.debug("Openshift CLI client isn't installed.")
        return False
    try:
        return bool(
            oc.selector(OSP_CONTROLPLANE, all_namespaces=True).objects())
    except oc.OpenShiftPythonException:
        return False


def get_dataplane_ssh_keypair():
    private_key = ""
    public_key = ""
    try:
        with oc.project(CONF.tobiko.podified.osp_project):
            secret_object = oc.selector(DP_SSH_SECRET_NAME).object()
        private_key = secret_object.as_dict()['data']['ssh-privatekey']
        public_key = secret_object.as_dict()['data']['ssh-publickey']
    except oc.OpenShiftPythonException as err:
        LOG.error("Error while trying to get Dataplane secret SSH Key: %s",
                  err)
    return private_key, public_key


def list_edpm_nodes():
    nodes = []
    with oc.project(CONF.tobiko.podified.osp_project):
        nodesets = oc.selector(OSP_DP_NODESET).objects()
    for nodeset in nodesets:
        nodeset_spec = nodeset.as_dict()['spec']
        nodeset_status = nodeset.as_dict()['status']
        node_template = nodeset_spec['nodeTemplate']
        nodeset_nodes = nodeset_spec['nodes']
        group_name = _get_group(nodeset_spec['services'])
        for node in nodeset_nodes.values():
            node_hostname = node.get('hostName')
            node_dict = {
                'hostname': node_hostname,
                'host': (node['ansible'].get('ansibleHost') or
                         _get_edpm_node_ctlplane_ip_from_status(
                             node_hostname, nodeset_status)),
                'group': group_name,
                'port': (
                    node.get('ansible', {}).get('ansiblePort') or
                    node_template.get('ansible', {}).get('ansiblePort')),
                'username': (
                    node.get('ansible', {}).get('ansibleUser') or
                    node_template.get('ansible', {}).get('ansibleUser')),
            }
            nodes.append(node_dict)
    return nodes


def list_ocp_workers():
    # oc.selector("nodes") does not need to run on a specific OCP project
    nodes_sel = oc.selector(OCP_WORKERS)
    ocp_workers = []
    for node in nodes_sel.objects():
        node_dict = node.as_dict()
        ocp_workers.append({
            'hostname': _get_ocp_worker_hostname(node_dict),
            'addresses': _get_ocp_worker_addresses(node_dict)
        })
    return ocp_workers


def power_on_edpm_node(nodename):
    _set_edpm_node_online_status(nodename, online=True)


def power_off_edpm_node(nodename):
    _set_edpm_node_online_status(nodename, online=False)


def _set_edpm_node_online_status(nodename, online):
    if _is_baremetal_crd_available() is False:
        LOG.info("BareMetal operator is not available in the deployment. "
                 "Starting and stopping EDPM nodes is not supported.")
        return
    try:
        with oc.project(CONF.tobiko.podified.osp_project):
            bm_node = oc.selector(f"{OSP_BM_HOST}/{nodename}").objects()[0]
    except oc.OpenShiftPythonException as err:
        LOG.info(f"Error while trying to get BareMetal Node '{nodename}' "
                 f"from Openshift. Error: {err}")
        return
    except IndexError:
        LOG.error(f"Node {nodename} not found in the {OSP_BM_HOST} CRs.")
        return
    bm_node.model.spec['online'] = online
    try:
        # NOTE(slaweq): returned status is 0 when all operations where
        # finished successfully. Otherwise status will be different than
        # 0, like in the shell scripts
        if not bool(bm_node.apply().status()):
            _wait_for_poweredOn_status(nodename, online)
    except oc.OpenShiftPythonException as err:
        LOG.error(f"Error while applying new online state: {online} for "
                  f"the node: {nodename}. Error: {err}")


def _wait_for_poweredOn_status(nodename, expected_status,
                               timeout: tobiko.Seconds = None):
    for attempt in tobiko.retry(
            timeout=timeout,
            count=10,
            interval=5.,
            default_timeout=30):
        LOG.debug(f"Checking power status of the '{nodename}'.")
        try:
            with oc.project(CONF.tobiko.podified.osp_project):
                poweredOn = oc.selector(
                    f"{OSP_BM_HOST}/{nodename}"
                ).objects()[0].model.status['poweredOn']
        except oc.OpenShiftPythonException as err:
            LOG.error("Error while trying to get 'poweredOn' state of "
                      f"the node {nodename}. Error: {err}")
        else:
            if poweredOn == expected_status:
                LOG.debug(f"Actual poweredOn state of the node {nodename} "
                          f"is: '{poweredOn}' which is as expected.")
                return True
            LOG.debug(f"Actual poweredOn state is: '{poweredOn}' != "
                      f" '{expected_status}'")
        attempt.check_limits()


def get_ovndbcluter(ovndbcluster_name):
    with oc.project(CONF.tobiko.podified.osp_project):
        ovndbcluter = oc.selector(
            f"{OVNDBCLUSTER}/{ovndbcluster_name}").objects()
    if len(ovndbcluter) != 1:
        tobiko.fail(f"Unexpected number of {OVNDBCLUSTER}/{ovndbcluster_name} "
                    f"objects obtained: {len(ovndbcluter)}")
    return ovndbcluter[0].as_dict()


def get_pods(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).objects()


def get_pod_names(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).qnames()


def get_pod_count(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).count_existing()


def delete_pods(labels=None):
    with oc.project(CONF.tobiko.podified.osp_project):
        return oc.selector('pods', labels=labels).delete()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os
import unittest

import netaddr
import pandas as pd
import testtools

from tobiko import config
from tobiko.openstack import keystone
from tobiko.openstack import metalsmith
from tobiko import tripleo
from tobiko.tripleo import pacemaker
from tobiko.tripleo import services
from tobiko.tripleo import processes
import tobiko

CONF = config.CONF


@tripleo.skip_if_missing_undercloud
class OvercloudKeystoneCredentialsTest(testtools.TestCase):

    def test_fetch_overcloud_credentials(self):
        env = tripleo.load_overcloud_rcfile()
        self.assertTrue(env['OS_AUTH_URL'])
        self.assertTrue(env.get('OS_USERNAME') or env.get('OS_USER_ID'))
        self.assertTrue(env['OS_PASSWORD'])
        self.assertTrue(env.get('OS_TENANT_NAME') or
                        env.get('OS_PROJECT_NAME') or
                        env.get('OS_TENANT_ID') or
                        env.get('OS_PROJECT_ID'))

    def test_overcloud_keystone_credentials(self):
        fixture = tripleo.overcloud_keystone_credentials()
        self.assertIsInstance(fixture,
                              keystone.KeystoneCredentialsFixture)
        credentials = keystone.keystone_credentials(fixture)
        credentials.validate()

    def test_overcloud_keystone_session(self):
        session = tripleo.overcloud_keystone_session()
        client = keystone.get_keystone_client(session=session)
        endpoints = keystone.list_endpoints(client=client)
        self.assertNotEqual([], endpoints)

    def test_overcloud_keystone_client(self):
        client = tripleo.overcloud_keystone_client()
        _services = keystone.list_services(client=client)
        self.assertTrue(_services)


@tripleo.skip_if_missing_overcloud
class OvercloudMetalsmithApiTest(testtools.TestCase):

    def test_list_overcloud_nodes(self):
        nodes = tripleo.list_overcloud_nodes()
        self.assertTrue(nodes)
        for node in nodes:
            node_ip = metalsmith.find_instance_ip_address(
                instance=node, check_connectivity=True)
            self.assertIsInstance(node_ip, netaddr.IPAddress)

    def test_find_overcloud_nodes(self):
        node = tripleo.find_overcloud_node()
        node_ip = metalsmith.find_instance_ip_address(instance=node,
                                                      check_connectivity=True)
        self.assertIsInstance(node_ip, netaddr.IPAddress)

    def test_get_overcloud_node_ip_address(self):
        overcloud_node_ip = tripleo.overcloud_node_ip_address()
        self.assertIsInstance(overcloud_node_ip, netaddr.IPAddress)

    def test_overcloud_host_config(self):
        instance = tripleo.find_overcloud_node()
        host_config = tobiko.setup_fixture(
            tripleo.overcloud_host_config(instance=instance))
        instance_ips = set()
        for ips in instance.ip_addresses().values():
            instance_ips.update(ips)
        self.assertIn(host_config.host, instance_ips)
        self.assertIsInstance(host_config.hostname, str)
        netaddr.IPAddress(host_config.hostname)
        self.assertEqual(CONF.tobiko.rhosp.ssh_port,
                         host_config.port)
        self.assertEqual(tripleo.get_overcloud_ssh_username(),
                         host_config.username)
        key_filename = tobiko.tobiko_config_path(
            CONF.tobiko.rhosp.ssh_key_filename)
        self.assertEqual(key_filename, host_config.key_filename)
        self.assertTrue(os.path.isfile(key_filename))
        self.assertTrue(os.path.isfile(key_filename + '.pub'))

    def test_overcloud_ssh_client_connection(self):
        instance = tripleo.find_overcloud_node()
        ssh_client = tripleo.overcloud_ssh_client(instance=instance)
        ssh_client.connect()


@tripleo.skip_if_missing_overcloud
class OvercloudPacemakerTest(testtools.TestCase):
    """
    Assert that all pacemaker resources are in
    healthy state
    """

    def test_get_pacemaker_resource_table(self):
        resource_table = pacemaker.get_pcs_resources_table()
        self.assertIsInstance(resource_table, pd.DataFrame)

    def test_pacemaker_resources_health(self):
        pcs_health = pacemaker.PacemakerResourcesStatus()
        self.assertTrue(pcs_health.all_healthy)


@tripleo.skip_if_missing_overcloud
class OvercloudServicesTest(testtools.TestCase):
    """
    Assert that a subset of overcloud services are in running state
    across the overcloud nodes
    """

    services_status = tobiko.required_fixture(
        services.OvercloudServicesStatus)

    def test_get_services_resource_table(self):
        self.assertIsInstance(self.services_status.oc_services_df,
                              pd.DataFrame)

    def test_overcloud_services(self):
        self.assertTrue(self.services_status.basic_overcloud_services_running)

    def test_get_overcloud_nodes_running_pcs_resource(self):
        nodes_list = pacemaker.get_overcloud_nodes_running_pcs_resource(
            resource_type='(ocf::heartbeat:rabbitmq-cluster):',
            resource_state='Started')
        self.assertIsInstance(nodes_list, list)


@tripleo.skip_if_missing_overcloud
class OvercloudProcessesTest(testtools.TestCase):
    """
    Assert that a subset of overcloud processes are in running state
    across the overcloud nodes
    """

    def test_get_processes_resource_table(self):
        ops = processes.OvercloudProcessesStatus()
        self.assertIsInstance(ops.oc_procs_df,
                              pd.DataFrame)

    def test_overcloud_processes(self):
        ops = processes.OvercloudProcessesStatus()
        self.assertTrue(ops.basic_overcloud_processes_running)


@tripleo.skip_if_missing_undercloud
class OvercloudVersionTest(unittest.TestCase):
    # TODO(eolivare): move the properties to a common class, since they are
    # duplicate in OvercloudVersionTest and UndercloudVersionTest

    @property
    def lower_version(self) -> str:
        v = tripleo.overcloud_version()
        if v.micro > 0:
            lower_v = f"{v.major}.{v.minor}.{v.micro - 1}"
        elif v.minor > 0:
            lower_v = f"{v.major}.{v.minor -1}.{v.micro}"
        elif v.major > 0:
            lower_v = f"{v.major -1}.{v.minor}.{v.micro}"
        else:
            raise ValueError(f"wrong version: {v}")
        return lower_v

    @property
    def same_version(self) -> str:
        v = tripleo.overcloud_version()
        return f"{v.major}.{v.minor}.{v.micro}"

    @property
    def higher_version(self) -> str:
        v = tripleo.overcloud_version()
        return f"{v.major}.{v.minor}.{v.micro + 1}"

    def test_overcloud_version(self):
        version = tripleo.overcloud_version()
        self.assertTrue(tobiko.match_version(version, min_version='13'))

    def test_has_overcloud(self):
        self.assertTrue(tripleo.has_overcloud())

    def test_has_overcloud_with_min_version(self):
        self.assertTrue(
            tripleo.has_overcloud(min_version=self.same_version))

    def test_has_overcloud_with_min_version_lower(self):
        self.assertTrue(
            tripleo.has_overcloud(min_version=self.lower_version))

    def test_has_overcloud_with_min_version_higher(self):
        self.assertFalse(
            tripleo.has_overcloud(min_version=self.higher_version))

    def test_has_overcloud_with_max_version(self):
        self.assertFalse(
            tripleo.has_overcloud(max_version=self.same_version))

    def test_has_overcloud_with_max_version_lower(self):
        self.assertFalse(
            tripleo.has_overcloud(max_version=self.lower_version))

    def test_has_overcloud_with_max_version_higher(self):
        self.assertTrue(
            tripleo.has_overcloud(max_version=self.higher_version))

    def test_skip_unless_has_overcloud(self):
        self._assert_test_skip_unless_has_overcloud_dont_skip()

    def test_skip_unless_has_overcloud_with_min_version(self):
        self._assert_test_skip_unless_has_overcloud_dont_skip(
            min_version=self.same_version)

    def test_skip_unless_has_overcloud_with_min_version_lower(self):
        self._assert_test_skip_unless_has_overcloud_dont_skip(
            min_version=self.lower_version)

    def test_skip_unless_has_overcloud_with_min_version_higher(self):
        self._assert_test_skip_unless_has_overcloud_skip(
            min_version=self.higher_version)

    def test_skip_unless_has_overcloud_with_max_version(self):
        self._assert_test_skip_unless_has_overcloud_skip(
            max_version=self.same_version)

    def test_skip_unless_has_overcloud_with_max_version_lower(self):
        self._assert_test_skip_unless_has_overcloud_skip(
            max_version=self.lower_version)

    def test_skip_unless_has_overcloud_with_max_version_higher(self):
        self._assert_test_skip_unless_has_overcloud_dont_skip(
            max_version=self.higher_version)

    def _assert_test_skip_unless_has_overcloud_dont_skip(
            self,
            min_version: tobiko.VersionType = None,
            max_version: tobiko.VersionType = None):
        executed = []

        @tripleo.skip_unless_has_overcloud(min_version=min_version,
                                           max_version=max_version)
        def decorated_function():
            executed.append(True)

        self.assertFalse(executed)
        decorated_function()
        self.assertTrue(executed)

    def _assert_test_skip_unless_has_overcloud_skip(
            self,
            min_version: tobiko.VersionType = None,
            max_version: tobiko.VersionType = None):
        @tripleo.skip_unless_has_overcloud(min_version=min_version,
                                           max_version=max_version)
        def decorated_function():
            raise self.fail('Not skipped')

        with self.assertRaises(unittest.SkipTest):
            decorated_function()



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.shell import sh


LOG = log.getLogger(__name__)


@keystone.skip_if_missing_service(name='octavia')
class OctaviaBasicTrafficScenarioTest(testtools.TestCase):
    """Octavia traffic scenario test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balanacer.
    """
    lb = None
    listener = None
    pool = None
    server_stack = tobiko.required_fixture(
        stacks.OctaviaServerStackFixture)
    other_server_stack = tobiko.required_fixture(
        stacks.OctaviaOtherServerStackFixture)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaBasicTrafficScenarioTest, self).setUp()

        self.lb, self.listener, self.pool = octavia.deploy_ipv4_amphora_lb(
            servers_stacks=[self.server_stack, self.other_server_stack]
        )

    def test_round_robin_traffic(self):
        _test_traffic(
            pool_id=self.pool.id,
            ip_address=self.lb.vip_address,
            lb_algorithm=self.pool.lb_algorithm,
            protocol=self.listener.protocol,
            port=self.listener.protocol_port)


def _test_traffic(pool_id: str, ip_address: str, lb_algorithm: str,
                  protocol: str, port: int):
    # For 5 minutes we ignore specific exceptions as we know
    # that Octavia resources are being provisioned
    for attempt in tobiko.retry(timeout=300.):
        try:
            octavia.check_members_balanced(pool_id=pool_id,
                                           ip_address=ip_address,
                                           lb_algorithm=lb_algorithm,
                                           protocol=protocol,
                                           port=port)
            break

        # TODO oschwart: change the following without duplicating code
        except octavia.RoundRobinException:
            if lb_algorithm == 'ROUND_ROBIN':
                LOG.exception(f"Traffic didn't reach all members after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
            if attempt.is_last:
                raise
        except (octavia.TrafficTimeoutError,
                sh.ShellCommandFailed):
            LOG.exception(f"Traffic didn't reach all members after "
                          f"#{attempt.number} attempts and "
                          f"{attempt.elapsed_time} seconds")
            if attempt.is_last:
                raise


@neutron.skip_unless_is_ovn()
@keystone.skip_if_missing_service(name='octavia')
class OctaviaOVNProviderTrafficTest(testtools.TestCase):
    """Octavia OVN provider traffic test.

    Create an OVN provider load balancer with 2 members that run a server
    application,
    Create a client that is connected to the load balancer VIP port via FIP,
    Generate TCP network traffic from the client to the load balancer FIP.
    """
    lb = None
    listener = None
    pool = None
    server_stack = tobiko.required_fixture(
        stacks.OctaviaServerStackFixture)
    other_server_stack = tobiko.required_fixture(
        stacks.OctaviaOtherServerStackFixture)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaOVNProviderTrafficTest, self).setUp()

        self.lb, self.listener, self.pool = octavia.deploy_ipv4_ovn_lb(
            servers_stacks=[self.server_stack, self.other_server_stack]
        )

    def test_source_ip_port_traffic(self):
        """Send traffic to the load balancer FIP to test source ip port
        """
        _test_traffic(
            pool_id=self.pool.id,
            ip_address=self.lb.vip_address,
            lb_algorithm=self.pool.lb_algorithm,
            protocol=self.listener.protocol,
            port=self.listener.protocol_port)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import contextlib
import random

from oslo_log import log
import pytest
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack.base import _fixture as base_fixture
from tobiko.shell import ping


LOG = log.getLogger(__name__)


@keystone.skip_unless_has_keystone_credentials()
class BaseServerTest(testtools.TestCase):
    __test__ = False  # forbid this test for being recollected
    stack: tobiko.RequiredFixture[stacks.ServerStackFixture]
    peer_stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    def test_1_server_name(self):
        server = self.ensure_server()
        self.assertEqual(self.stack.server_name, server.name)

    @pytest.mark.server_shutoff
    def test_2_shutoff_server(self):
        # ensure boot is finished before turning VM down
        self.ensure_server(status='ACTIVE')
        self.assert_is_reachable()

        self.ensure_server(status='SHUTOFF')
        self.assert_is_unreachable()

    @pytest.mark.server_active
    def test_3_activate_server(self):
        # ensure server is down before activating it
        self.ensure_server(status='SHUTOFF')
        self.assert_is_unreachable()

        self.ensure_server(status='ACTIVE')
        self.stack.wait_for_guest_os_ready()
        self.assert_is_reachable()

    @pytest.mark.server_migrate
    @nova.skip_if_missing_hypervisors(count=2)
    def test_4_migrate_server(self):
        self._test_migrate_server(live=False)

    @pytest.mark.server_migrate
    @nova.skip_if_missing_hypervisors(count=2)
    def test_5_live_migrate_server(self):
        self._test_migrate_server(live=True)

    @pytest.mark.server_migrate
    @nova.skip_if_missing_hypervisors(count=3)
    def test_6_migrate_server_with_host(self):
        """Tests cold migration actually ends on target hypervisor
        """
        self._test_migrate_server_with_host(live=False)

    @pytest.mark.server_migrate
    @nova.skip_if_missing_hypervisors(count=3)
    def test_7_live_migrate_server_with_host(self):
        self._test_migrate_server_with_host(live=True)

    def _test_migrate_server(self,
                             live: bool,
                             block_migration: bool = None):
        """Tests cold migration actually changes hypervisor
        """
        server = self.ensure_server(status='ACTIVE')
        initial_hypervisor = nova.get_server_hypervisor(server)

        server = self.migrate_server(live=live,
                                     block_migration=block_migration)
        final_hypervisor = nova.get_server_hypervisor(server)
        self.assertNotEqual(initial_hypervisor, final_hypervisor)

    def _test_migrate_server_with_host(self,
                                       live: bool):
        """Tests cold migration actually ends on target hypervisor
        """
        server = self.ensure_server(status='ACTIVE')
        initial_hypervisor = nova.get_server_hypervisor(server)

        hypervisors = nova.list_hypervisors(
            status='enabled', state='up').select(
            lambda h: h.hypervisor_hostname != initial_hypervisor)
        if not hypervisors:
            tobiko.skip_test("Cannot find a valid hypervisor host to migrate "
                             "server to")
        target_hypervisor = random.choice(hypervisors).hypervisor_hostname

        server = self.migrate_server(host=target_hypervisor,
                                     live=live)
        final_hypervisor = nova.get_server_hypervisor(server)
        self.assertNotEqual(initial_hypervisor, final_hypervisor)
        self.assertEqual(target_hypervisor, final_hypervisor)

    def ensure_server(self, status: str = None):
        server = self.stack.server_details
        if status not in [None, server.status]:
            LOG.debug(f"Ensuring server '{server.id}' status changes:\
                      '{server.status}' -> '{status}'")
            assert isinstance(status, str)
            server = self.stack.ensure_server_status(status)
            self.assertEqual(status, server.status)
        return server

    def migrate_server(self,
                       live=False,
                       host: str = None,
                       block_migration: bool = None) \
            -> nova.NovaServer:
        self.assert_is_reachable()
        with self.handle_migration_errors():
            server = self.stack.migrate_server(live=live,
                                               host=host,
                                               block_migration=block_migration)
        self.assert_is_reachable()
        return server

    @contextlib.contextmanager
    def handle_migration_errors(self):
        try:
            yield
        except (nova.PreCheckMigrateServerError,
                nova.NoValidHostFoundMigrateServerError,
                nova.NotInLocalStorageMigrateServerError,
                nova.NotInSharedStorageMigrateServerError) as ex:
            self.skipTest(str(ex))

    def assert_is_reachable(self):
        self.stack.assert_is_reachable()
        ping.assert_reachable_hosts(self.stack.list_fixed_ips(),
                                    timeout=300.,
                                    ssh_client=self.peer_stack.ssh_client)
        if self.stack.has_vlan:
            self.stack.assert_vlan_is_reachable()

    def assert_is_unreachable(self):
        self.stack.assert_is_unreachable()
        ping.assert_unreachable_hosts(self.stack.list_fixed_ips(),
                                      timeout=300.,
                                      ssh_client=self.peer_stack.ssh_client)
        if self.stack.has_vlan:
            self.stack.assert_vlan_is_unreachable()


class CirrosServerStackFixture(stacks.CirrosServerStackFixture):

    def validate_created_stack(self):
        stack = super().validate_created_stack()
        server = nova.get_server(self.server_id)
        if server.status != 'SHUTOFF':
            if server.status != 'ACTIVE':
                try:
                    nova.activate_server(server)
                except nova.WaitForServerStatusTimeout as ex:
                    raise base_fixture.InvalidFixtureError(
                        name=self.stack_name) from ex
        return stack


@pytest.mark.minimal
class CirrosServerTest(BaseServerTest):
    __test__ = True
    stack = tobiko.required_fixture(CirrosServerStackFixture)


class UbuntuServerStackFixture(stacks.UbuntuServerStackFixture):
    pass


class UbuntuServerTest(BaseServerTest):
    __test__ = True
    stack = tobiko.required_fixture(UbuntuServerStackFixture)
